from typing import Optional, Literal, List, Union
import numpy as np

from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData
from ... import _constants as constants


def rank_genes_groups(
    adata: ConnectorAnnData,
    groupby: str,
    groups: Union[List[str], str],
    reference: Union[List[str], str] = "rest",
    method: Literal["venice", "t-test", "wilcoxon"] = "venice",
    normalize_method: Optional[
        Literal[
            constants.NormalizeMethod.RAW,
            constants.NormalizeMethod.LOG1P_NORMALIZE,
            constants.NormalizeMethod.SQRT_NORMALIZE,
        ]
    ] = constants.NormalizeMethod.LOG1P_NORMALIZE,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Spatial Region Segmentation.

    Parameters
    ----------
    method: `Literal["venice", "t-test", "wilcoxon"]`, Default: "venice"
        Test statistic method.
    normalize_method: `Literal["raw", "log1p-normalized", "sqrt-normalized"]`, Default: "log1p-normalized"
        Using raw expression or scaling (log1p, square root) for test statistic.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    metadata = adata.obs[groupby].values

    if isinstance(groups, str):
        groups = [groups]
    group_1_indices = np.where(np.isin(metadata, groups))[0]
    if len(group_1_indices) == 0:
        raise ValueError(f"Not found any cells in {groups}.")
    group_1_name = str(groups)

    if reference == "rest":
        group_2_indices = np.where(~np.isin(metadata, groups))[0]
        group_2_name = "rest"
    else:
        if isinstance(reference, str):
            reference = [reference]
        group_2_indices = np.where(np.isin(metadata, reference))[0]
        group_2_name = str(reference)

    data_id_1 = data_id_2 = adata._extend_information[constants.ConnectorKeys.DATA_ID.value]

    analysis = adata.self_init(Analysis)
    analysis.de.differential_expression_genes(
        data_id_1=data_id_1,
        data_id_2=data_id_2,
        group_1_indices=group_1_indices.tolist(),
        group_2_indices=group_2_indices.tolist(),
        group_1_name=group_1_name,
        group_2_name=group_2_name,
        method=method,
        normalize_method=normalize_method,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )



import os
import uuid
import logging
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from functools import cached_property
from typing import List, Iterable, Collection, Tuple, Union, Optional

import zarr
import numpy as np
import pandas as pd
from scipy import sparse
from anndata import AnnData
from anndata._io.specs import read_elem
from anndata._core.file_backing import AnnDataFileManager
from anndata._core.aligned_mapping import AxisArrays, Layers, PairwiseArrays

from . import _constants as constants
from ._api import ZarrElement, OpenAPI, PyAPI
from ._analysis import Analysis


DEFAULT_LOGGER = logging.getLogger("spatialx_sdks_stdout")
SLICE = Union[int, slice, Collection]
SLICE_CHECK = (int, slice, Collection)


class AnnDataElement(ZarrElement):
    @cached_property
    def _data(self):
        return {}

    @cached_property
    def _parent(self):
        return self.self_init(ConnectorAnnData, os.path.dirname(self._path))

    def __setitem__(self, key: str, value):
        return super().__setitem__(key, value)

    def _get_mem_item(self, key: str):
        return self._data[key]

    def _get_item_iplm(self, key: str):
        return self.self_init(ZarrElement, path=os.path.join(self._path, key))

    def __getitem__(self, key: str):
        if key in self._data:
            return self._get_mem_item(key)

        key = self.get_id_by_name(key)
        return self._get_item_iplm(key)

    def keys(self):
        return np.unique(
            [k for k in sorted(self._name2id.keys()) if not k.startswith("AN-")] + list(self._data.keys())
        )

    def update(self):
        data = self._data
        new_obj: AnnDataElement = self.self_init(type(self))
        new_obj._data.update(data)
        return new_obj


class ConnectorDataFrame(AnnDataElement, AxisArrays):
    @cached_property
    def index_key(self):
        return self.attrs.get(
            constants.AnnDataKeys.INDEX_KEY.value,
            constants.AnnDataKeys.INDEX_KEY.value,
        )

    @property
    def index_name(self):
        return self.index_key

    @cached_property
    def columns(self) -> List[str]:
        return [col for col in self.keys() if col != self.index_key]

    @cached_property
    def index(self):
        return pd.Index(read_elem(self.group[self.index_key]), name=self.index_name)

    def _get_mem_item(self, key: str):
        return pd.Series(
            self._data[key],
            index=self.index,
            name=self.get_name_by_id(key),
        )

    def _get_item_iplm(self, key: str):
        return pd.Series(
            read_elem(self.group[key]),
            index=self.index,
            name=self.get_name_by_id(key),
        )

    def __getitem__(self, keys: Union[str, List[str]]) -> pd.DataFrame:
        if isinstance(keys, list):
            return pd.concat([self[key] for key in keys], axis=1, keys=keys)
        return super().__getitem__(keys)

    def to_df(self) -> pd.DataFrame:
        """Convert to pandas DataFrame"""
        keys = [key for key in self.keys() if key != constants.AnnDataKeys.INDEX_KEY.value]
        if keys:
            return self[keys]
        return pd.DataFrame(index=self.index)

    @property
    def loc(self):
        return self.to_df().loc

    @property
    def iloc(self):
        return self.to_df().iloc


class ConnectorObs(ConnectorDataFrame):
    @property
    def _axis(self):
        return constants.AnnDataAxis.OBS.value

    @property
    def index_name(self):
        return "Barcodes"


class ConnectorVar(ConnectorDataFrame):
    @property
    def _axis(self):
        return constants.AnnDataAxis.VAR.value

    @property
    def index_name(self):
        return "Genes"


class ConnectorEmbeddings(AnnDataElement, AxisArrays):
    def _get_item_iplm(self, key: str):
        return read_elem(self.group[key])


class ConnectorObsm(ConnectorEmbeddings):
    @property
    def _axis(self):
        return constants.AnnDataAxis.OBS.value


class ConnectorVarm(ConnectorEmbeddings):
    @property
    def _axis(self):
        return constants.AnnDataAxis.VAR.value


class ConnectorPairwiseArrays(AnnDataElement, PairwiseArrays):
    def _get_item_iplm(self, key: str):
        return read_elem(self.group[key])


class ConnectorObsp(ConnectorPairwiseArrays):
    @property
    def _axis(self):
        return constants.AnnDataAxis.OBS.value


class ConnectorVarp(ConnectorPairwiseArrays):
    @property
    def _axis(self):
        return constants.AnnDataAxis.VAR.value


@dataclass
class ArrayChunk:
    index: int
    start: int
    end: int
    data: np.ndarray


class ChunkReader:
    _array: zarr.Array = None
    _cur_chunk: ArrayChunk = None

    def __init__(self, array: zarr.Array, chunk_sz: int):
        self._array = array
        self._chunk_sz = chunk_sz

    def get_chunk(self, i: int) -> ArrayChunk:
        if self._cur_chunk is None or self._cur_chunk.index != i:
            s = i * self._chunk_sz
            e = min(len(self._array), s + self._chunk_sz)
            self._cur_chunk = ArrayChunk(index=i, start=s, end=e, data=self._array[s:e])
        return self._cur_chunk

    def read(self, start: int, end: int) -> np.ndarray:
        sz = end - start
        out = np.empty(sz, dtype=self._array.dtype)
        s_chunk = start // self._chunk_sz
        e_chunk = (end - 1) // self._chunk_sz

        for i in range(s_chunk, e_chunk + 1):
            cur_chunk = self.get_chunk(i)
            out_s = max(0, cur_chunk.start - start)
            out_e = min(sz, cur_chunk.end - start)
            chunk_s = max(0, start - cur_chunk.start)
            chunk_e = min(cur_chunk.end, end) - cur_chunk.start
            out[out_s:out_e] = cur_chunk.data[chunk_s:chunk_e]
        return out

    def __getitem__(self, i: slice) -> np.ndarray:
        return self.read(i.start, i.stop)


class ConnectorCSCMatrix(AnnDataElement, sparse.csc_matrix):
    @property
    def data(self) -> zarr.Array:
        return self.group["data"]

    @property
    def indices(self) -> zarr.Array:
        return self.group["indices"]

    @cached_property
    def indptr(self) -> zarr.Array:
        return self.group["indptr"][:]

    @cached_property
    def shape(self) -> tuple:
        return self.attrs["shape"]

    @property
    def major_axis(self):
        return 1

    @property
    def minor_axis(self):
        return 0

    def to_memory(self) -> sparse.csc_matrix:
        return sparse.csc_matrix(
            (self.data[:], self.indices[:], self.indptr[:]),
            shape=self.shape,
            dtype=self.data.dtype,
        )

    @cached_property
    def _readers(self) -> Tuple[ChunkReader, ChunkReader]:
        chunk_sz = max(
            0 if self.data.chunks is None else self.data.chunks[0],
            0 if self.indices.chunks is None else self.indices.chunks[0],
        )
        data_reader = ChunkReader(self.data, chunk_sz) if chunk_sz else self.data
        indices_reader = ChunkReader(self.indices, chunk_sz) if chunk_sz else self.indices
        return data_reader, indices_reader

    def __bulking_indices(self, indices: Collection) -> Tuple[Collection[Tuple[int, int]], np.ndarray]:
        seps = (np.diff(indices) != 1).nonzero()[0]
        indptr = np.empty(len(seps) + 2, dtype=np.int64)
        indptr[0] = 0
        indptr[1:-1] = seps + 1
        indptr[-1] = len(indices)
        segments = [
            (indices[s], indices[e - 1] + 1)
            for s, e in zip(indptr[:-1], indptr[1:])
        ]
        return segments, indptr

    def __get_index_from_slice(
        self, i: Union[SLICE, Tuple[SLICE, SLICE]], axis: int, max_len: int,
    ) -> Optional[np.ndarray]:
        ii = None
        if isinstance(i, Tuple) and len(i) > axis:
            ii = i[axis]
        elif isinstance(i, SLICE_CHECK) and axis == 0:
            ii = i

        if np.isscalar(ii):
            return np.array([ii], dtype=np.int32)
        if isinstance(ii, slice) and ii != slice(None):
            return np.arange(
                start=ii.start or 0,
                stop=ii.stop or max_len,
                step=ii.step,
                dtype=np.int32,
            )
        if isinstance(ii, Collection):
            return np.array(ii, dtype=np.int32)

        return None

    def _slice_major_axis(self, index: Iterable[int]) -> sparse.csc_matrix:
        indptr = np.empty(len(index) + 1, dtype=self.indptr.dtype)
        indptr[0] = 0
        indptr[1:] = np.diff(self.indptr)[index]
        indptr = np.cumsum(indptr, dtype=self.indptr.dtype)
        indices = np.empty(indptr[-1], dtype=self.indices.dtype)
        data = np.empty(indptr[-1], dtype=self.data.dtype)
        data_reader, indices_reader = self._readers
        query_slices, query_indptr = self.__bulking_indices(index)
        for i, query_slice in enumerate(query_slices):
            dst_s = indptr[query_indptr[i]]
            dst_e = indptr[query_indptr[i + 1]]
            if dst_e == dst_s:
                continue
            src_s = self.indptr[query_slice[0]]
            src_e = self.indptr[query_slice[1]]
            data[dst_s: dst_e] = data_reader[src_s: src_e]
            indices[dst_s: dst_e] = indices_reader[src_s: src_e]

        shape = (self.shape[self.minor_axis], len(index))
        return sparse.csc_matrix((data, indices, indptr), shape=shape, dtype=data.dtype)

    def _slice_both_axes(self, col_idx: Iterable[int], row_idx: Iterable[int]) -> sparse.csc_matrix:
        data_arrs, indices_arrs = [], []
        indptr = np.zeros(len(col_idx) + 1, dtype=self.indptr.dtype)

        mask = np.zeros(self.shape[self.minor_axis], dtype=bool)
        mask[row_idx] = True

        data_reader, indices_reader = self._readers
        for i, pos in enumerate(col_idx):
            s, e = self.indptr[pos], self.indptr[pos + 1]
            slice_data = data_reader[s:e]
            slice_indices = indices_reader[s:e]

            filtered_idx = mask[slice_indices]
            data_arrs.append(slice_data[filtered_idx])
            indices_arrs.append(slice_indices[filtered_idx])
            indptr[i + 1] = len(data_arrs[-1])

        indptr = np.cumsum(indptr)
        data = np.concatenate(data_arrs)
        indices = np.concatenate(indices_arrs)

        indices_mapping = np.zeros(self.shape[self.minor_axis], dtype=self.indices.dtype)
        indices_mapping[row_idx] = np.arange(len(row_idx), dtype=self.indices.dtype)
        indices = indices_mapping[indices]

        shape = (len(row_idx), len(col_idx))
        return sparse.csc_matrix((data, indices, indptr), shape=shape, dtype=data.dtype)

    def __getitem__(self, i: Union[SLICE, Tuple[SLICE, SLICE]]) -> sparse.csc_matrix:
        if not isinstance(i, tuple) and not isinstance(i, SLICE_CHECK):
            raise ValueError("Invalid slice")

        major_len = self.shape[self.major_axis]
        minor_len = self.shape[self.minor_axis]
        major_idx = self.__get_index_from_slice(i, self.major_axis, major_len)
        minor_idx = self.__get_index_from_slice(i, self.minor_axis, minor_len)

        if major_idx is None and minor_idx is None:
            return self.to_memory()

        if minor_idx is None:
            return self._slice_major_axis(major_idx)

        if len(minor_idx) != len(set(minor_idx)):
            raise ValueError("Minor axis slice on sparse matrix must be unique")

        major_idx = np.arange(self.shape[self.major_axis], dtype=np.int32) \
            if major_idx is None else major_idx

        res = self._slice_both_axes(major_idx, minor_idx)

        if len(major_idx) == 1 and len(minor_idx) == 1:
            res = res.toarray()[0][0]

        return res


class ConnectorAnnData(ZarrElement, AnnData):
    @cached_property
    def X(self) -> ConnectorCSCMatrix:
        return self.self_init(
            ConnectorCSCMatrix,
            os.path.join(self._path, constants.AnnDataAttr.X.value)
        )

    @cached_property
    def obs_names(self):
        return self.obs.index

    @cached_property
    def n_obs(self):
        return len(self.obs_names)

    @cached_property
    def var_names(self):
        return self.var.index

    @cached_property
    def n_vars(self):
        return len(self.var_names)

    def __repr__(self) -> str:
        descr = (
            f"{type(self).__name__} object with "
            f"n_obs × n_vars = {self.n_obs} × {self.n_vars}"
        )
        for attr in constants.AnnDataAttr:
            if attr.value not in self.group:
                continue
            if attr == constants.AnnDataAttr.X:
                continue
            keys = getattr(self, attr.value, {}).keys()
            if len(keys) > 0:
                descr += f"\n    {attr.value}: {str(list(keys))[1:-1]}"
        return descr

    def to_memory(self) -> AnnData:
        return AnnData(
            X=self.X.to_memory(),
            obs=self.obs.to_df(),
            var=self.var.to_df(),
            obsm={k: self.obsm[k] for k in self.obsm.keys()},
            varm={k: self.varm[k] for k in self.varm.keys()},
            obsp=self.obsp,
            varp=self.varp,
            layers=self.layers,
            uns=self.uns,
        )

    def copy(self) -> AnnData:
        return self.to_memory()

    @cached_property
    def _obs(self) -> ConnectorObs:
        return self.self_init(
            ConnectorObs,
            os.path.join(self._path, constants.AnnDataAttr.OBS.value)
        )

    @cached_property
    def _var(self) -> ConnectorVar:
        return self.self_init(
            ConnectorVar,
            os.path.join(self._path, constants.AnnDataAttr.VAR.value)
        )

    @cached_property
    def _obsm(self) -> ConnectorObsm:
        return self.self_init(
            ConnectorObsm,
            os.path.join(self._path, constants.AnnDataAttr.OBSM.value)
        )

    @cached_property
    def _varm(self) -> ConnectorVarm:
        return self.self_init(
            ConnectorVarm,
            os.path.join(self._path, constants.AnnDataAttr.VARM.value)
        )

    @cached_property
    def _obsp(self):
        return self.self_init(
            ConnectorObsp,
            os.path.join(self._path, constants.AnnDataAttr.OBSP.value)
        )

    @cached_property
    def _varp(self):
        return self.self_init(
            ConnectorVarp,
            os.path.join(self._path, constants.AnnDataAttr.VARP.value)
        )

    @cached_property
    def _raw(self):
        return None

    @cached_property
    def _layers(self):
        DEFAULT_LOGGER.warning("Create `layers` in memory.")
        return Layers(self)

    @cached_property
    def _uns(self):
        DEFAULT_LOGGER.warning("Create `uns` in memory.")
        return {}

    def _sanitize(self):
        DEFAULT_LOGGER.warning("[SKIP] Not support `_sanitize` function.")

    @property
    def filename(self):
        return None

    @cached_property
    def file(self):
        return AnnDataFileManager(self, None)

    @property
    def is_view(self):
        return False

    def upload_metadata(self, metadata: Union[pd.DataFrame, str, List[str]]):
        if isinstance(metadata, str):
            metadata = [metadata]
        if isinstance(metadata, list):
            metadata = self.obs[metadata]

        openapi: OpenAPI = self.self_init(OpenAPI)
        pyapi: PyAPI = self.self_init(PyAPI)
        with TemporaryDirectory(dir=".") as temp_dir:
            filename = f"{uuid.uuid4().hex}.tsv.gz"
            path = os.path.join(temp_dir, filename)
            metadata.to_csv(path, sep="\t")
            res = openapi.upload_file(
                file_path=path,
                folder_name="upload_metadata",
                upload_id=uuid.uuid4().hex,
            )
            csv_path = res[constants.ConnectorKeys.PATH.value]
            pyapi.import_metadata(
                study_id=self._extend_information[constants.ConnectorKeys.STUDY_ID.value],
                sample_id=self._extend_information[constants.ConnectorKeys.SAMPLE_ID.value],
                table_id=self._extend_information[constants.ConnectorKeys.TABLE_ID.value],
                csv_path=csv_path,
            )
        self.update()

    def upload_embeddings(self, embeddings: Union[str, List[str]]):
        if isinstance(embeddings, str):
            embeddings = [embeddings]
        adata = AnnData(
            X=sparse.csc_matrix((self.n_obs, 1)),
            obs=pd.DataFrame(index=self.obs_names),
            obsm={k: self.obsm[k] for k in embeddings}
        )

        openapi: OpenAPI = self.self_init(OpenAPI)
        analysis: Analysis = self.self_init(Analysis)
        with TemporaryDirectory(dir=".") as temp_dir:
            filename = f"{uuid.uuid4().hex}.h5ad"
            path = os.path.join(temp_dir, filename)
            adata.write_h5ad(path, compression="gzip")
            res = openapi.upload_file(
                file_path=path,
                folder_name="upload_embeddings",
                upload_id=uuid.uuid4().hex,
            )
            h5ad_path = res[constants.ConnectorKeys.PATH.value]
            analysis.embeddings.upload_embeddings(h5ad_path, info_args=self._extend_information)
        self.update()

    def update(self):
        self._obs = self._obs.update()
        self._obsm = self._obsm.update()
        self._obsp = self._obsp.update()
        self._var = self._var.update()
        self._varm = self._varm.update()
        self._varp = self._varp.update()



from typing import Optional

from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData


def region_segmentation(
    adata: ConnectorAnnData,
    radius: float,
    mpp: float,
    species: str,
    resolution: float = 0.5,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Spatial Region Segmentation.

    Parameters
    ----------
    radius: `float`
        Radius (micrometer) to build spatial graph
    mpp: `float`
        Micrometers per pixel.
    species: `str`
        Species of data.
    resolution : ``float``, Default: 0.5
        Resolution parameter, higher lead to more communities.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.spatial_analysis.region_segmentation(
        radius=radius,
        mpp=mpp,
        species=species,
        resolution=resolution,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def neighborhood_analysis(
    adata: ConnectorAnnData,
    radius: float,
    mpp: float,
    annotation_key: str = "",
    deconvolution_key: str = "",
    tau: Optional[float] = None,
    n_centroids: int = 30,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Spatial Neighborhood Analysis.

    Parameters
    ----------
    radius: `float`
        Radius (micrometer) to build spatial graph
    mpp: `float`
        Micrometers per pixel.
    annotation_key: `str`
        Metadata of cell-type to perform cell-type neighborhood analysis
    deconvolution_key: `str`
        Metadata of deconvolution_key cell-type to perform cell-type neighborhood analysis
    tau : `Optional[float]`
        Facilitating fine-tuned modulation of the impact of neighboring cells.
    n_centroids: `int`
        Number of centroids to run hierarchical clustering.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.spatial_analysis.neighborhood_analysis(
        radius=radius,
        mpp=mpp,
        annotation_key=annotation_key,
        deconvolution_key=deconvolution_key,
        tau=tau,
        n_centroids=n_centroids,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()



from ._embeddings import umap, tsne
from ._clustering import louvain, leiden, kmeans
from ._annotations import metadata_reference, deconvolution, list_deconvolution_models
from ._de import rank_genes_groups
from ._spatial import region_segmentation, neighborhood_analysis


__ALL__ = [
    umap,
    tsne,
    louvain,
    leiden,
    kmeans,
    metadata_reference,
    deconvolution,
    list_deconvolution_models,
    rank_genes_groups,
    region_segmentation,
    neighborhood_analysis,
]



from typing import Literal, Optional

from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData


def umap(
    adata: ConnectorAnnData,
    embedding_key: str,
    n_neighbors: int = 15,
    local_connectivity: float = 1,
    init: Literal["pca", "spectral", "random"] = "pca",
    deterministic: bool = True,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Uniform Manifold Approximation and Projection for Dimension Reduction.

    Parameters
    ----------
    embedding_key: `str`
        Embedding to run UMAP, should be result of PCA or scVI.
    n_neighbors : Optional[`int`], Default: None
        The number of nearest neighbors.
    local_connectivity : `float`, Default: 1.0
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.
    init : `str`, Default: "pca"
        How to initialize the low dimensional embedding. Options are:
        * `"spectral"` : use a spectral embedding of the fuzzy 1-skeleton (non-deterministic).
        * `"random"` : assign initial embedding positions at random.
        * `"pca"` : pca low dimensions of input matrix.
    deterministic : `bool`, Default: True
        Making sure with same initialization, results will remains the same across runs on the same device.
        Deterministic version will converge a lot slower on big datasets (hence lower quality on same setting).
        If you prefer quality/performance, disable `deterministic`
        Note: `"spectral"` initialization is non-deterministic across runs.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.embeddings.umap(
        embedding_key=embedding_key,
        title=title,
        n_neighbors=n_neighbors,
        local_connectivity=local_connectivity,
        init=init,
        deterministic=deterministic,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def tsne(
    adata: ConnectorAnnData,
    embedding_key: str,
    n_neighbors: int = 90,
    perplexity: float = 30,
    local_connectivity: float = 1,
    init: Literal["pca", "gaussian", "input"] = "pca",
    title: Optional[str] = None,
    **kwargs,
):
    """
    FFT-accelerated Interpolation-based t-SNE implementation.

    Parameters
    ----------
    embedding_key: `str`
        Embedding to run UMAP, should be result of PCA or scVI.
    n_neighbors : Optional[`int`], Default: None
        The number of nearest neighbors.
    perplexity : ``float``, default = ``30``
        The perplexity is related to the number of nearest neighbors that is used in
        other manifold learning algorithms.
        Larger datasets usually require a larger perplexity.
        Consider selecting a value between ``5`` and ``50``.
        Different values can result in significantly different results.
    local_connectivity : `float`, Default: 1.0
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.
    init : `str`, Default: "pca"
        How to initialize the low dimensional embedding. Options are:
        * ``"gaussian"`` : init using gaussian noise.
        * ``"input"`` : 2 first row input matrix.
        * ``"pca"`` : pca 2 first row input matrix.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.embeddings.tsne(
        embedding_key=embedding_key,
        n_neighbors=n_neighbors,
        perplexity=perplexity,
        local_connectivity=local_connectivity,
        init=init,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()



from typing import Optional

from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData


def louvain(
    adata: ConnectorAnnData,
    embedding_key: str,
    resolution: float,
    n_neighbors: int = 90,
    local_connectivity: float = 1,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Louvain clustering implementation.

    Parameters
    ----------
    embedding_key: `str`
        Embedding to run UMAP, should be result of PCA or scVI.
    resolution : ``float``
        Resolution parameter, higher lead to more communities.
    n_neighbors : Optional[`int`], Default: None
        The number of nearest neighbors.
    local_connectivity : `float`, Default: 1.0
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.clustering.louvain(
        embedding_key=embedding_key,
        resolution=resolution,
        n_neighbors=n_neighbors,
        local_connectivity=local_connectivity,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def leiden(
    adata: ConnectorAnnData,
    embedding_key: str,
    resolution: float,
    n_neighbors: int = 90,
    local_connectivity: float = 1,
    title: Optional[str] = None,
    **kwargs,
):
    """
    Leiden clustering implementation.

    Parameters
    ----------
    embedding_key: `str`
        Embedding to run UMAP, should be result of PCA or scVI.
    resolution : ``float``
        Resolution parameter, higher lead to more communities.
    n_neighbors : Optional[`int`], Default: None
        The number of nearest neighbors.
    local_connectivity : `float`, Default: 1.0
        The local connectivity required -- i.e. the number of nearest
        neighbors that should be assumed to be connected at a local level.
        The higher this value the more connected the manifold becomes
        locally. In practice this should be not more than the local intrinsic
        dimension of the manifold.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.clustering.leiden(
        embedding_key=embedding_key,
        resolution=resolution,
        n_neighbors=n_neighbors,
        local_connectivity=local_connectivity,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def kmeans(
    adata: ConnectorAnnData,
    embedding_key: str,
    n_clusters: int,
    title: Optional[str] = None,
    **kwargs,
):
    """
    k-means clustering implementation.

    Parameters
    ----------
    embedding_key: `str`
        Embedding to run UMAP, should be result of PCA or scVI.
    n_clusters: ``int``
        Number of clusters.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.clustering.kmeans(
        embedding_key=embedding_key,
        n_clusters=n_clusters,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()



import os
import logging
from typing import List, Optional, Union
from requests.auth import HTTPBasicAuth

import zarr

from . import _constants as constants
from ._utils import time_cache
from ._api import Connector, ZarrElement, OpenAPI, PyAPI
from ._anndata import ConnectorAnnData


DEBUG_MODE = os.environ.get("DEBUG_MODE", "") == "TRUE"
DEFAULT_LOGGER = logging.getLogger("spatialx_sdks_stdout")


def transform_images(images, transformation):
    from spatial_image import SpatialImage
    from dask import array as dask_array
    from spatialdata import transform
    from spatialdata.transformations._utils import _set_transformations

    spimages = SpatialImage(
        dask_array.from_array(images["0"][:]),
        dims=("c", "y", "x"),
    )
    _set_transformations(spimages, {"trans": transformation})
    return transform(spimages, to_coordinate_system="trans")


def get_images(images: zarr.Group, resolution_level: int = 3):
    from spatialdata.transformations import (
        Affine,
        Scale,
        Sequence,
        Translation,
        Identity,
    )
    from spatialdata.transformations._utils import _set_transformations

    data = images[resolution_level]

    transformation = Identity()
    if "btr_global_transformation" in images.attrs:
        transformation = Affine(
            images.attrs["btr_global_transformation"],
            input_axes=("x", "y", "z"),
            output_axes=("x", "y", "z"),
        ).to_affine(
            input_axes=("x", "y"),
            output_axes=("x", "y"),
        )
    elif "scalefactors" in images.attrs or "translation" in images.attrs:
        transformation = []
        if "scalefactors" in images.attrs:
            transformation.append(
                Scale(
                    [images.attrs["scalefactors"], images.attrs["scalefactors"]],
                    axes=("x", "y"),
                )
            )
        if "scalefactors" in images.attrs:
            transformation.append(
                Translation(images.attrs["translation"], axes=("x", "y"))
            )
        transformation = Sequence(transformation).to_affine(
            input_axes=("x", "y"),
            output_axes=("x", "y"),
        )

    transformation = Sequence([
        Scale([2 ** resolution_level, 2 ** resolution_level], axes=("x", "y")),
        transformation,
        Scale([1 / 2 ** resolution_level, 1 / 2 ** resolution_level], axes=("x", "y"))
    ]).to_affine(
        input_axes=("x", "y"),
        output_axes=("x", "y"),
    )

    data = transform_images(data, transformation)
    transformation = Sequence([
        Translation(data.transform["transformation"].translation, axes=("x", "y")),
        Scale([2 ** resolution_level, 2 ** resolution_level], axes=("x", "y")),
    ]).to_affine(
        input_axes=("x", "y"),
        output_axes=("x", "y"),
    )
    _set_transformations(data, {"origin": transformation})
    return data


class Images(ZarrElement):
    def __getitem__(self, key: str):
        return self.group[self.get_id_by_name(key)]

    def get_images(self, key: str, resolution_level: int = 3):
        return get_images(self[key], resolution_level=resolution_level)

    def get_channel_names(self, elem: Union[str, zarr.Group]):
        if isinstance(elem, str):
            elem = self[elem]
        for attr in elem.attrs.asdict().values():
            if isinstance(attr, dict) and constants.SpatialAttrs.CHANNELS_KEY.value in attr:
                return [
                    name[constants.SpatialAttrs.CHANNEL_LABEL.value]
                    for name in attr[constants.SpatialAttrs.CHANNELS_KEY.value]
                ]
        return []

    def set_channel_names(self, elem: Union[str, zarr.Group], channel_names: List[str]):
        if isinstance(elem, str):
            elem = self[elem]
        pyapi: PyAPI = self.self_init(PyAPI)
        return pyapi.rename_image_channels(
            self._extend_information[constants.ConnectorKeys.STUDY_ID.value],
            self._extend_information[constants.ConnectorKeys.SAMPLE_ID.value],
            self.get_id_by_name(elem.path),
            channel_names,
        )


class Shapes(ZarrElement):
    def __getitem__(self, key: str):
        import numpy as np
        from geopandas import GeoDataFrame
        from spatialdata._io.io_shapes import (
            ShapesFormats,
            ShapesFormatV01,
            _parse_version,
            from_ragged_array,
            _get_transformations_from_ngff_dict,
            _set_transformations,
        )

        def _read_shapes(store: zarr.Group) -> GeoDataFrame:
            """Read shapes from a zarr store."""
            f = zarr.open(store, mode="r")
            attrs = f.attrs.asdict()
            version = _parse_version(f, expect_attrs_key=True)
            assert version is not None
            format = ShapesFormats[version]

            if isinstance(format, ShapesFormatV01):
                coords = np.array(f["coords"][:])
                index = np.array(f["Index"][:])
                typ = format.attrs_from_dict(attrs)
                if typ.name == "POINT":
                    radius = np.array(f["radius"][:])
                    geometry = from_ragged_array(typ, coords)
                    geo_df = GeoDataFrame({"geometry": geometry, "radius": radius}, index=index)
                else:
                    offsets = []
                    for i in range(3):
                        try:
                            offsets.append(np.array(f[f"offset{i}"][:]).flatten())
                        except:
                            break
                    geometry = from_ragged_array(typ, coords, offsets)
                    geo_df = GeoDataFrame({"geometry": geometry}, index=index)
            else:
                raise ValueError(
                    f"Unsupported shapes format {format} from version {version}. "
                    "Please update the spatialdata library."
                )

            transformations = _get_transformations_from_ngff_dict(attrs["coordinateTransformations"])
            _set_transformations(geo_df, transformations)

            if constants.ConnectorKeys.MPP.value in attrs:
                geo_df[constants.ConnectorKeys.MPP.value] = attrs[constants.ConnectorKeys.MPP.value]

            return geo_df

        return _read_shapes(
            self._open_zarr_root(os.path.join(self._path, self.get_id_by_name(key)))
        )

    def simplify(self, key: str, tolerance: Optional[float]):
        pyapi: PyAPI = self.self_init(PyAPI)
        return pyapi.simplify_segmentation(
            self._extend_information[constants.ConnectorKeys.STUDY_ID.value],
            self._extend_information[constants.ConnectorKeys.SAMPLE_ID.value],
            self.get_id_by_name(key),
            tolerance,
        )


class Points(ZarrElement):
    def __getitem__(self, key: str):
        import io
        import requests

        import zarr
        import pandas as pd
        import dask
        from dask import dataframe as dd
        from spatialdata._io.io_points import (
            DaskDataFrame,
            PointsFormats,
            _parse_version,
            _get_transformations_from_ngff_dict,
            _set_transformations,
        )

        @dask.delayed
        def parquet_from_http(url, headers=None, meta=None):
            result = requests.get(url, headers=headers)
            df = pd.read_parquet(io.BytesIO(result.content))
            if meta is not None:
                for k, v in meta.items():
                    df[k] = v
            return df

        key = self.get_id_by_name(key)
        group = zarr.open(self._open_zarr_root(os.path.join(self._path, key)))
        attrs = group.attrs.asdict()
        root_path = os.path.join(
            self._get_zarr_path(os.path.join(self._path, key)),
            "points.parquet",
        )

        paths = []
        divisions = tuple(attrs["divisions"])
        genes = list(attrs.get("gene_to_partition", {}).keys())
        for i in range(len(divisions) - 1):
            paths.append(os.path.join(root_path, f"part.{i}.parquet"))

        delayed_download = [
            parquet_from_http(
                path,
                headers=self._headers,
                meta=None if not genes else dict(gene=genes[i]),
            )
            for i, path in enumerate(paths)
        ]
        points: DaskDataFrame = dd.from_delayed(delayed_download, divisions=divisions)

        version = _parse_version(group, expect_attrs_key=True)
        assert version is not None

        transformations = _get_transformations_from_ngff_dict(attrs["coordinateTransformations"])
        _set_transformations(points, transformations)

        if constants.ConnectorKeys.MPP.value in attrs:
            points[constants.ConnectorKeys.MPP.value] = attrs[constants.ConnectorKeys.MPP.value]

        p_attrs = PointsFormats[version].attrs_from_dict(attrs)
        if len(p_attrs):
            points.attrs["spatialdata_attrs"] = p_attrs

        return points


class Tables(ZarrElement):
    def __getitem__(self, key: str):
        key = self.get_id_by_name(key)
        return self.self_init(
            ConnectorAnnData,
            path=os.path.join(self._path, key),
            extend_information={constants.ConnectorKeys.TABLE_ID.value: key},
        )


class SpatialData(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        data_id: Optional[str] = None,
        sample_id: Optional[str] = None,
        verify_ssl: bool = False,
        authentication: Optional[HTTPBasicAuth] = None,
    ):
        openapi = OpenAPI(domain, token, verify_ssl=verify_ssl, authentication=authentication)
        if sample_id is not None:
            details = openapi.get_sample_detail(sample_id)[constants.ConnectorKeys.SAMPLE.value]
            self.elements = None
        elif data_id is not None:
            details = openapi.get_sample_data_detail(data_id)

            submitted_elements: dict = details.get("map_submit_result", {})
            elements: List[str] = []
            for value in submitted_elements.values():
                if not isinstance(value, dict):
                    continue
                for v in value.values():
                    elements.extend(v.keys() if isinstance(v, dict) else [v])
            self.elements = elements

        self.study_id = details[constants.ConnectorKeys.STUDY_ID.value]
        self.sample_id = details[constants.ConnectorKeys.SAMPLE_ID.value]
        self.data_id = data_id
        study_path = openapi.get_study_detail(
            self.study_id
        )[constants.ConnectorKeys.STUDY.value][constants.ConnectorKeys.DATA_PATH.value]
        self._path = f"{study_path}/spatial/{self.sample_id}"

        super().__init__(
            domain, token,
            verify_ssl=verify_ssl,
            authentication=authentication,
        )

    def _init_attribute(self, Attribute, key: str):
        return self.self_init(
            Attribute,
            path=os.path.join(self._path, key),
            elements=self.elements,
            extend_information={
                constants.ConnectorKeys.DATA_ID.value: self.data_id,
                constants.ConnectorKeys.STUDY_ID.value: self.study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: self.sample_id,
            },
        )

    @property
    @time_cache(5)
    def images(self) -> Images:
        return self._init_attribute(Images, "images")

    @property
    @time_cache(5)
    def shapes(self) -> Shapes:
        return self._init_attribute(Shapes, "shapes")

    @property
    @time_cache(5)
    def points(self) -> Points:
        return self._init_attribute(Points, "points")

    @property
    @time_cache(5)
    def tables(self) -> Tables:
        return self._init_attribute(Tables, "table")

    def __repr__(self):
        obj_str = f"{type(self).__name__} with elements:"
        for key in ["images", "shapes", "points", "tables"]:
            element: ZarrElement = getattr(self, key)
            if len(element.keys()) == 0:
                continue
            obj_str += "\n\t" + element.repr()
        return obj_str



from typing import List, Dict, Union, Optional, Literal, Tuple
import os
import json
from enum import Enum
from tqdm import tqdm
from pathlib import Path
from requests.auth import HTTPBasicAuth
from functools import singledispatchmethod

from ._constants import SpatialAttrs
from ._constants import DefaultGroup
from ._constants import ConnectorKeys
from ._constants import SubmissionType
from ._constants import ExpressionSubmission
from ._constants import SubmissionElementKeys

from ._api import PyAPI
from ._api import OpenAPI
from ._api import Connector
from ._analysis import Analysis
from ._spatialdata import SpatialData
from ._anndata import ConnectorAnnData
from ._studies import Studies, Study, Data
from ._utils import get_chunk_size, format_print


class SpatialXConnector(Connector):
    """
    SpatialX Connector
    Supporting to work with spatial data via notebook.
    """

    def __init__(
        self,
        domain: str,
        token: str,
        verify_ssl: bool = False,
        authentication: Optional[HTTPBasicAuth] = None,
    ):
        """
        Construct parameters for train and query k-nearest neighbors

        Parameters
        ----------
        domain: ``str``
            SpatialX domain
        token: ``str``
            User's token
        verify_ssl: ``bool``, default: False
            Verify SSL or not.
        """
        super().__init__(domain, token, verify_ssl=verify_ssl, authentication=authentication)
        self.__pyapi = self.self_init(PyAPI)
        self.__openapi = self.self_init(OpenAPI)
        self.__analysis = self.self_init(Analysis)

    @property
    def pyapi(self) -> PyAPI:
        return self.__pyapi

    @property
    def openapi(self) -> OpenAPI:
        return self.__openapi

    @property
    def analysis(self) -> Analysis:
        return self.__analysis

    @property
    def info(self):
        """Current user's information"""
        info = self.openapi.info
        return {
            field: info[field]
            for field in ConnectorKeys.INFORMATION_FIELDS.value
        }

    @property
    def groups(self):
        """List all reachable groups of current user in domain server."""
        group_info = self.openapi.groups
        groups = {
            v: k for k, v in group_info[ConnectorKeys.DEFAULT.value].items()
        }
        for group in group_info[ConnectorKeys.GROUPS.value]:
            groups[group["name"]] = group["id"]
        return groups

    @property
    def external_folders(self):
        """List all reachable mounted shared folders of current user from BBrowserX/BioStudio."""
        mounts = self.openapi.mounts
        return {
            folder["name"]: folder["path"]
            for folder in mounts["folders"]
        } if mounts.get("folders", None) else {}

    @property
    def external_s3(self):
        """List all reachable mounted s3 of current user from BBrowserX/BioStudio."""
        mounts = self.openapi.mounts
        return mounts.get("s3", {})

    @property
    def folders(self):
        """List all reachable mounted shared folders of current user in domain server."""
        defaults = {
            folder["name"]: folder["path"]
            for folder in self.openapi.info["default_mount"]["folders"]
        }
        return dict(self.external_folders.items() | defaults.items())

    @property
    def s3(self):
        """List all reachable mounted s3 clouds of current user in domain server."""
        s3_buckets = {}
        for s3 in self.openapi.info["default_mount"]["s3"]:
            name = s3.get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = s3["id"]
            s3_buckets[name] = s3["path"]

        for external_s3 in self.external_s3:
            name = external_s3.get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = external_s3["id"]
            if name in s3_buckets:
                name = f"{name} - External AWS"
            s3_buckets[name] = external_s3["path"]

        for internal_s3 in self.openapi.s3:
            name = internal_s3["map_settings"].get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = internal_s3["map_settings"]["id"]
            if name in s3_buckets:
                name = f"{name} - Internal AWS"
            s3_buckets[name] = internal_s3["map_settings"]["path"]

        return s3_buckets

    def listdir(
        self,
        path: str,
        ignore_hidden: bool = True,
        get_details: bool = False,
    ) -> Union[List[Dict[str, Union[str, int, dict]]], List[str]]:
        """
        List all files and folders with path in domain server

        Parameters
        ----------
        path: ``str``
            path of folder to list
        ignore_hidden: ``bool``, default: True
            Ignore hidden files/folders or not
        get_details: ``bool``, default: False
            Get details information or not

        Returns
        -------
        results: ``Union[List[Dict[str, Union[str, int, dict]]], List[str]]``
            Folders and files with their information
        """
        dir_elements = self.openapi.list_dir(
            path, ignore_hidden=ignore_hidden
        )[ConnectorKeys.ENTITIES.value]
        if get_details:
            return dir_elements
        return [element[ConnectorKeys.NAME.value] for element in dir_elements]

    def parse_data_information(
        self,
        data_name: str,
        technology: str,
        data_path: str,
        args: dict = None,
        min_genes: int = 1,
        min_counts: int = 1,
        mito_controls_percentage: float = 0.25,
    ) -> dict:
        """
        Parse information of data to valid format for submission

        Parameters
        ----------
        data_name: ``str``
            Name of spatial data
        technology: ``str``
            Technology of spatial data
        data_path: ``str``
            Path to spatial data

        Returns
        -------
        submission_data: ``dict``
            Auto-detect information for submission
        """
        info = self.pyapi.parse_data_information(data_name, technology, data_path)
        if args is not None:
            args = [
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in args.items()
            ]
            for sample_data in info:
                sample_data[ConnectorKeys.ARGS.value] = sample_data[ConnectorKeys.ARGS.value] + args

        for sample_data in info:
            if sample_data[ConnectorKeys.KWARGS.value] is None:
                sample_data[ConnectorKeys.KWARGS.value] = []
            sample_data[ConnectorKeys.KWARGS.value] = sample_data[ConnectorKeys.KWARGS.value] + [
                {ConnectorKeys.KEY.value: key, ConnectorKeys.VALUE.value: value}
                for key, value in dict(
                    min_genes=min_genes,
                    min_counts=min_counts,
                    mito_controls_percentage=mito_controls_percentage,
                ).items()
            ]

        return info

    def parse_multiple_samples_information(
        self,
        technology: str,
        data_path: str,
        sample_name_mapping: dict = {},
        data_name_mapping: dict = {},
        args: dict = None,
        min_genes: int = 1,
        min_counts: int = 1,
        mito_controls_percentage: float = 0.25,
    ) -> List[dict]:
        """
        Parse information of multiple samples to valid format for submission

        Parameters
        ----------
        technology: ``str``
            Technology of spatial data
        data_path: ``str``
            Path to spatial data

        Returns
        -------
        submission_samples: ``dict``
            Auto-detect information for submission
        """
        if args is not None:
            args = [
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in args.items()
            ]

        results: List[dict] = []
        for folder in self.listdir(data_path, get_details=True):
            if folder[ConnectorKeys.TYPE.value] != ConnectorKeys.DIRECTORY.value:
                continue
            try:
                info = self.pyapi.parse_data_information(
                    data_name_mapping.get(folder[ConnectorKeys.NAME.value], folder[ConnectorKeys.NAME.value]),
                    technology,
                    os.path.join(data_path, folder[ConnectorKeys.NAME.value]),
                )
                if args is not None:
                    for sample_data in info:
                        sample_data[ConnectorKeys.ARGS.value] = sample_data[ConnectorKeys.ARGS.value] + args

                for sample_data in info:
                    if sample_data.get(ConnectorKeys.KWARGS.value, None) is None:
                        sample_data[ConnectorKeys.KWARGS.value] = []
                    sample_data[ConnectorKeys.KWARGS.value] = sample_data[ConnectorKeys.KWARGS.value] + [
                        {ConnectorKeys.KEY.value: key, ConnectorKeys.VALUE.value: value}
                        for key, value in dict(
                            min_genes=min_genes,
                            min_counts=min_counts,
                            mito_controls_percentage=mito_controls_percentage,
                        ).items()
                    ]

                results.append({
                    ConnectorKeys.SAMPLE_NAME.value: sample_name_mapping.get(
                        folder[ConnectorKeys.NAME.value], folder[ConnectorKeys.NAME.value]
                    ),
                    ConnectorKeys.DATA.value: info,
                })
            except Exception as e:
                print(f"Fail to parse data in {folder[ConnectorKeys.NAME.value]}: {e}")

        return results

    def list_study(
        self, group: str, species: str, **kwargs
    ) -> Studies:
        """
        List reachable studies

        Parameters
        ----------
        group: ``str``
            Group of studies
        species: ``str``
            Species of studies

        Returns
        -------
        results: ``Studies``
            List of studies and their information

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        if isinstance(group, Enum):
            group = group.value
        group = self.groups.get(group, group)
        return self.self_init(
            Studies,
            data=self.openapi.list_study(group, species, **kwargs)["list"],
        )

    def get_study_detail(self, study_id: str, **kwargs) -> Study:
        """
        Get details information of study

        Parameters
        ----------
        study_id: ``str``
            Id of study

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of study

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_study_detail(study_id, **kwargs)

    def list_sample(self, study_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        List samples in a study

        Parameters
        ----------
        study_id: ``str``
            Id of study

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            List of samples and their information

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.list_sample(study_id, **kwargs)["list"]

    def get_sample_detail(self, sample_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        Get details information of sample data

        Parameters
        ----------
        sample_id: ``str``
            Id of sample

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of data

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_sample_detail(sample_id, **kwargs)

    def rename_sample(self, sample_id: str, new_name: str, **_):
        """
        Rename sample

        Parameters
        ----------
        sample_id: ``str``
            Id of sample of rename
        new_name: ``str``
            New name of sample
        """
        return self.openapi.rename_sample(sample_id, new_name)

    def delete_sample(self, sample_id: str, **_):
        """
        Delete sample

        Parameters
        ----------
        sample_id: ``str``
            Id of sample of delete
        """
        return self.openapi.delete_sample(sample_id)

    def get_sample_data_detail(self, data_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        Get details information of sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of data

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_sample_data_detail(data_id, **kwargs)

    def get_sample_data_elements(self, data_id: str) -> Dict[str, List[str]]:
        """
        Get elements of sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data

        Returns
        -------
        results: ``Dict[str, List[str]]``
            Elements of data
        """
        results: Dict[str, Dict[str, dict]] =  \
            self.openapi.get_sample_data_detail(data_id).get("map_submit_result", None)
        spdata = self.self_init(SpatialData, data_id=data_id)
        if results is None:
            return {}
        elements: Dict[str, List[str]] = {}
        for value in results.values():
            if not isinstance(value, dict):
                continue
            for k, v in value.items():
                if k not in elements:
                    elements[k] = []
                elements[k].extend(
                    getattr(spdata, SpatialAttrs.SPATIAL_ELEMENT_MAPPING.value[k]).get_name_by_id(
                        [v] if isinstance(v, str) else v.keys()
                    )
                )
        return elements

    def rename_sample_data(self, data_id: str, new_name: str, **_):
        """
        Rename sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data to rename
        new_name: ``str``
            New name of data
        """
        return self.openapi.rename_sample_data(data_id, new_name)

    def delete_sample_data(self, data_id: str, **_):
        """
        Delete sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data to delete
        """
        return self.openapi.delete_sample_data(data_id)

    def get_anndata(self, data_id: str, anndata_id: str) -> ConnectorAnnData:
        return self.self_init(SpatialData, data_id=data_id).tables[anndata_id]

    def add_sample(
        self,
        study_id: str,
        sample_name: str,
        sample_data: List[dict] = [],
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Add a sample to a existed study

        Parameters
        ----------
        study_id: ``str``
            Id of study
        sample_name: ``str``
            Sample name
        sample_data: ``List[dict]``, default: []
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """

        return self.openapi.create_sample(study_id, sample_name, sample_data)

    def add_sample_data(
        self,
        study_id: str,
        sample_id: str,
        sample_data: List[dict],
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Add data to a existed sample

        Parameters
        ----------
        study_id: ``str``
            Id of sample's root study
        sample_id: ``str``
            Id of sample
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """
        return self.openapi.add_sample_data(study_id, sample_id, sample_data)

    def create_study(self, group: str, species: str, title: str):
        """
        Create new study.

        Parameters
        ----------
        group: ``str``
            Group of study
        species: ``str``
            Species of data in study
        title: ``str``
            Title of study

        Returns
        -------
        study_id: `str`
            ID of the new study
        """
        if isinstance(group, Enum):
            group = group.value
        return self.openapi.create_study(self.groups.get(group, group), species, title)[ConnectorKeys.STUDY_ID.value]

    def __tracking_log_add_sample(self, res: dict, max_time: int = 600):
        data_path = self.get_study_detail(
            res[ConnectorKeys.STUDY_ID.value]
        )[ConnectorKeys.STUDY.value][ConnectorKeys.DATA_PATH.value]

        if ConnectorKeys.SUBMIT_ID.value in res:
            submission_id = res[ConnectorKeys.SUBMIT_ID.value]
            self.tracking_log(
                os.path.join(data_path, f"submit/{submission_id}/submission.log"), max_time=max_time
            )
            return

        for data in res[ConnectorKeys.SAMPLE_DATA.value]:
            submission_id = data[ConnectorKeys.SUBMIT_ID.value]
            self.tracking_log(
                os.path.join(data_path, f"submit/{submission_id}/submission.log"), max_time=max_time
            )

    def submit(
        self,
        group: str,
        species: str,
        title: str,
        sample_name: str,
        sample_data: List[dict],
        tracking: bool = False,
        max_time: int = 600,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        group: ``str``
            Group of study
        species: ``str``
            Species of data in study
        title: ``str``
            Title of study
        sample_name: ``str``
            Sample name
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """
        study_id = self.create_study(group, species, title)
        res = self.add_sample(study_id, sample_name, sample_data)
        if tracking:
            self.__tracking_log_add_sample(res, max_time)
        return res

    def add_multiple_samples(
        self,
        study_id: str,
        sample_data: List[dict],
    ) -> List[Dict[str, Union[str, List[dict]]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        study_id: ``str``
            ID of study to add samples.
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_multiple_samples_information`` function

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        results = []
        for sample in sample_data:
            sample_name = sample[ConnectorKeys.SAMPLE_NAME.value]
            data = sample[ConnectorKeys.DATA.value]
            results.append(self.add_sample(study_id, sample_name, data))

        return results

    def submit_multiple_samples(
        self,
        group: str,
        species: str,
        title: str,
        sample_data: List[dict],
    ) -> List[Dict[str, Union[str, List[dict]]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        group: ``str``
            Group of study
        species: ``str``
            Species of data in study
        title: ``str``
            Title of study
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_multiple_samples_information`` function

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        study_id = self.create_study(group, species, title)[ConnectorKeys.STUDY_ID.value]
        return self.add_multiple_samples(study_id, sample_data)

    def add_custom_sample(
        self,
        study_id: str,
        sample_name: str,
        data_name: str,
        technology: str,
        adding_types: List[str],
        paths: Dict[str, str] = {},
        args: Dict[str, str] = {},
        kwargs: Dict[str, str] = {},
        tracking: bool = False,
        max_time: int = 600,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        title: ``str``
            Title of elements
        study_id: ``str``
            Study ID
        sample_name: ``str``
            Sample Name
        data_name: `str`
            Sample Data Name
        technology: `str`
            Technology of data
        adding_types: `List[str]`
            Type of element to adding, defined in [
                `spatialx_connector.ImagesSubmission`,
                `spatialx_connector.SegmentationSubmission`,
                `spatialx_connector.TrasncriptsSubmission`,
                `spatialx_connector.ExpressionSubmission`,
            ]
        paths: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and their paths.
        args: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and values.

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        res = self.add_sample(
            study_id,
            sample_name,
            [
                dict(
                    name=data_name,
                    submission_type=SubmissionType.detect_submission_type(technology),
                    technology=technology,
                    identities=adding_types,
                    files=[
                        {
                            ConnectorKeys.KEY.value: key,
                            ConnectorKeys.VALUE.value: value,
                        }
                        for key, value in paths.items()
                    ],
                    folders=[],
                    args=[
                        {
                            ConnectorKeys.KEY.value: key,
                            ConnectorKeys.VALUE.value: value,
                        }
                        for key, value in args.items()
                    ] + [
                        {
                            # Ignore default types of technologies.
                            ConnectorKeys.KEY.value: "ignore_technology_elements",
                            ConnectorKeys.VALUE.value: True,
                        }
                    ],
                    kwargs=[
                        {
                            ConnectorKeys.KEY.value: key,
                            ConnectorKeys.VALUE.value: value,
                        }
                        for key, value in kwargs.items()
                    ],
                )
            ]
        )
        if tracking:
            self.__tracking_log_add_sample(res, max_time)
        return res

    def add_sample_data_element(
        self,
        title: str,
        data_id: str,
        adding_types: List[str],
        paths: Dict[str, str] = {},
        args: Dict[str, str] = {},
        tracking: bool = False,
        max_time: int = 600,
        **_,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        title: ``str``
            Title of elements
        data_id: ``str``
            Data ID
        adding_types: `List[str]`
            Type of element to adding, defined in [
                `spatialx_connector.ImagesSubmission`,
                `spatialx_connector.SegmentationSubmission`,
                `spatialx_connector.TrasncriptsSubmission`,
                `spatialx_connector.ExpressionSubmission`,
            ]
        paths: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and their paths.
        args: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and values.

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        data: Data = self.self_init(Data, data=data_id)
        res = data.add_elements(
            title,
            adding_types=adding_types,
            paths=paths,
            args=args,
        )
        if tracking:
            self.__tracking_log_add_sample(res, max_time)
        return res

    @singledispatchmethod
    def combine_multiple_datasets(
        self,
        study_id: str,
        data_name: str,
        technology: str,
        reference_anndata_paths: List[Tuple[str, str, str]],
        included_images: Optional[Dict[str, str]] = None,
        included_segmentation: Optional[Dict[str, str]] = None,
        metadata_action: Literal["ignore", "combine", "separate"] = "ignore",
    ):
        """
        Combine multiple datasets into one.

        Parameters
        ----------
        study_id: ``str``
            Study ID
        data_name: ``str``
            Name of the new combined dataset
        technology: ``str``
            Technology of the combined dataset
        reference_anndata_paths: ``List[Tuple[str, str, str]]``
            Information of multiple datasets, list of tuple (sample name, sample id, table name)
        included_images: ``Dict[str, str]``
            Mapping of table and its images
        included_segmentation: ``Dict[str, str]``
            Mapping of table and its segmentation
        metadata_action: ``Literal["ignore", "combine", "separate"]``
            What we do with multiple metadata
        """
        args = [
            {
                ConnectorKeys.KEY.value: SubmissionElementKeys.REFERENCE_ANNDATA_PATHS,
                ConnectorKeys.VALUE.value: json.dumps(reference_anndata_paths),
            },
            {
                ConnectorKeys.KEY.value: SubmissionElementKeys.METADATA_ACTION,
                ConnectorKeys.VALUE.value: metadata_action,
            },
            {
                # Ignore default types of technologies.
                ConnectorKeys.KEY.value: "ignore_technology_elements",
                ConnectorKeys.VALUE.value: True,
            }
        ]
        if included_images is not None and len(included_images) > 0:
            args.append(
                {
                    ConnectorKeys.KEY.value: SubmissionElementKeys.INCLUDED_IMAGES,
                    ConnectorKeys.VALUE.value: json.dumps(included_images),
                }
            )
        if included_segmentation is not None and len(included_segmentation) > 0:
            args.append(
                {
                    ConnectorKeys.KEY.value: SubmissionElementKeys.INCLUDED_SEGMENTATION,
                    ConnectorKeys.VALUE.value: json.dumps(included_segmentation),
                }
            )

        return self.add_sample(
            study_id=study_id,
            sample_name="Default combined samples",
            sample_data=[
                dict(
                    name=data_name,
                    submission_type="COMBINED_DATASETS",
                    technology=technology,
                    identities=[ExpressionSubmission.MERGE_MULTIPLE_ANNDATA],
                    files=[],
                    folders=[],
                    args=args,
                    kwargs=[],
                )
            ]
        )

    @combine_multiple_datasets.register(list)
    def _combine_multiple_datasets(
        self,
        data_ids: List[str],
        data_name: str,
        technology: str,
        metadata_action: Literal["ignore", "combine", "separate"] = "ignore",
    ):
        reference_anndata_paths = []
        included_images = {}
        included_segmentation = {}
        study_id = ""

        for data_id in data_ids:
            elements = self.get_sample_data_elements(data_id)
            if SpatialAttrs.ANNOTATED_DATA.value not in elements:
                continue

            details = self.get_sample_data_detail(data_id)
            study_id = details[ConnectorKeys.STUDY_ID.value]
            sample_id = details[ConnectorKeys.SAMPLE_ID.value]
            sample_name = self.get_sample_detail(sample_id)[ConnectorKeys.SAMPLE.value][ConnectorKeys.TITLE.value]
            reference_anndata_paths.append((sample_name, sample_id, elements[SpatialAttrs.ANNOTATED_DATA.value][0]))
            if SpatialAttrs.IMAGES_ELEMENT.value in elements:
                included_images[reference_anndata_paths[-1][-1]] = elements[SpatialAttrs.IMAGES_ELEMENT.value][0]
            if SpatialAttrs.SEGMENTATION_ELEMENT.value in elements:
                included_segmentation[reference_anndata_paths[-1][-1]] = \
                    elements[SpatialAttrs.SEGMENTATION_ELEMENT.value][0]

        return self.combine_multiple_datasets(
            study_id,
            data_name,
            technology,
            reference_anndata_paths,
            included_images=included_images,
            included_segmentation=included_segmentation,
            metadata_action=metadata_action,
        )

    def upload_file(
        self,
        file_path: str,
        server_folder_name: str = "",
        upload_id: str = "",
        is_chunk: bool = False,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        upload a small file

        Parameters
        ----------
        file_path: ``str``
            File location
        server_folder_name: ``str``
            Folder location in spatialx server
        upload_id: ``str``
            Upload ID

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Upload information
        """
        return self.openapi.upload_file(
            file_path=file_path,
            folder_name=server_folder_name,
            upload_id=upload_id,
            is_chunk=is_chunk,
        )

    def upload_big_file(
        self,
        file_path: str,
        chunk_size: int = 0,
        debug_mode: bool = False,
        server_folder_name: str = "",
        chunk_resp: dict = {},
        move_to_parent: bool = True,
        is_chunk: bool = False,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Upload a big file

        Parameters
        ----------
        file_path: ``str``
            File location
        chunk_size: ``int``
            Chunk size (bytes), 0: auto
        debug_mode: ``bool``
            Debug mode
        server_folder_name: ``str``
            Folder location in spatialx server

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Upload information
        """
        if not os.path.isfile(file_path):
            raise Exception(f"Invalid file: {file_path}")

        file_size = os.stat(os.path.abspath(file_path)).st_size
        upload_id = ""
        resp = chunk_resp
        if ConnectorKeys.UNIQUE_ID.value in resp:
            upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

        # Direct upload if small file
        if file_size < ConnectorKeys.UPLOAD_CHUNK_SMALL_SIZE.value:
            if ConnectorKeys.UNIQUE_ID.value in resp:
                upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

            return self.upload_file(
                file_path=file_path,
                server_folder_name=server_folder_name,
                upload_id=upload_id,
                is_chunk=is_chunk,
            )

        file_name = Path(file_path).name
        item_chunk_size = get_chunk_size(chunk_size, file_size)

        if (len(resp.keys()) == 0) or (len(upload_id) == 0):
            resp = self.openapi.upload_chunk_start(
                folder_name=server_folder_name,
                parent_is_file=2,
            )

            if ConnectorKeys.UNIQUE_ID.value in resp:
                upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

        file = open(file_path, "rb")
        file.seek(0, 0)
        sending_index = 0
        offset_size = 0
        progress_bar = None
        if debug_mode:
            progress_bar = tqdm(total=file_size, unit="B", unit_scale=True)

        while True:
            data = file.read(item_chunk_size)
            if not data:
                break

            offset_size = offset_size + item_chunk_size
            offset_size = min(file_size, offset_size)

            if debug_mode:
                format_print(f"Upload {file_path}, chunk index : {sending_index + 1} ...")

            if ConnectorKeys.ROOT_FOLDER.value in resp:
                resp_path = resp[ConnectorKeys.ROOT_FOLDER.value]
            else:
                resp_path = resp[ConnectorKeys.PATH.value]
            self.openapi.upload_chunk_process(
                chunk_size=item_chunk_size,
                file_size=file_size,
                offset=offset_size,
                file_name=file_name,
                folder_name=server_folder_name,
                upload_id=upload_id,
                path=resp_path,
                sending_index=sending_index,
                parent_is_file=2,
                file_data=data,
            )

            if debug_mode:
                if progress_bar is not None:
                    progress_bar.update(len(data))

            sending_index = sending_index + 1

        total_index = sending_index
        file.close()

        if ConnectorKeys.ROOT_FOLDER.value in resp:
            resp_path = resp[ConnectorKeys.ROOT_FOLDER.value]
        else:
            resp_path = resp[ConnectorKeys.PATH.value]
        resp2 = self.openapi.upload_chunk_merge(
            total_chunk=total_index,
            file_name=file_name,
            folder_name=server_folder_name,
            upload_id=upload_id,
            path=resp_path,
            parent_is_file=2,
            move_to_parent=move_to_parent,
        )

        if move_to_parent:
            return resp2
        return resp

    def upload_folder(
        self,
        dir_path: str,
        chunk_size: int = 0,
        debug_mode: bool = False,
        server_folder_name: str = "",
        chunk_resp: dict = {},
        is_chunk: bool = False,
    ) -> bool:
        """
        Upload folder as: zarr

        Parameters
        ----------
        dir_path: ``str``
            Folder location
        chunk_size: ``int``
            Chunk size (bytes), 0: auto
        debug_mode: ``bool``
            Debug mode
        server_folder_name: ``str``
            Folder location in spatialx server
        """
        if not Path(dir_path).is_dir():
            raise Exception(f"Invalid directory: {dir_path}.")

        root_folder_path = os.path.basename(dir_path)
        if server_folder_name and len(server_folder_name) > 0:
            root_folder_path = os.path.join(server_folder_name, root_folder_path)

        src_path = Path(dir_path)
        resp = chunk_resp

        for src_child in src_path.iterdir():
            if src_child.is_dir():
                dst_child = os.path.join(dir_path, src_child.name)
                self.upload_folder(
                    dir_path=dst_child,
                    chunk_size=chunk_size,
                    debug_mode=debug_mode,
                    server_folder_name=root_folder_path,
                    chunk_resp=resp,
                    is_chunk=True,
                )
            else:
                if src_child.is_symlink():
                    continue

                dst_child = os.path.join(dir_path, src_child.name)
                resp = self.upload_big_file(
                    file_path=dst_child,
                    chunk_size=chunk_size,
                    debug_mode=debug_mode,
                    server_folder_name=root_folder_path,
                    chunk_resp=resp,
                    move_to_parent=False,
                    is_chunk=True,
                )

        if is_chunk:
            return {}
        return self.openapi.upload_folder_finish(
            root_folder_path,
            resp[ConnectorKeys.UNIQUE_ID.value],
        )

    def list_lens_bulk_studies(self, host: str, token: str, group: str, species: str):
        from bioturing_connector.lens_bulk_connector import LensBulkConnector  # type: ignore

        connector = LensBulkConnector(host=host, token=token, ssl=True)
        connector.test_connection()
        if group == DefaultGroup.PERSONAL_WORKSPACE.value:
            group_id = DefaultGroup.LENS_GROUP_ID_PERSONAL_WORKSPACE.value
        elif group == DefaultGroup.ALL_MEMBERS.value:
            group_id = DefaultGroup.LENS_GROUP_ID_ALL_MEMBERS.value
        else:
            group_id = self.groups.get(group, group)

        studies_info = connector.get_all_studies_info_in_group(group_id=group_id, species=species)
        studies_info = [
            {
                **info,
                ConnectorKeys.SPECIES.value: species,
                ConnectorKeys.GROUP_ID.value: self.groups.get(
                    group, DefaultGroup.PERSONAL_WORKSPACE.value
                ),
            }
            for info in studies_info
        ]
        return studies_info

    def list_lens_sc_studies(self, host: str, token: str, group: str, species: str):
        from bioturing_connector.lens_sc_connector import LensSCConnector  # type: ignore

        connector = LensSCConnector(host=host, token=token, ssl=True)
        connector.test_connection()
        if isinstance(group, Enum):
            group = group.value
        if group == DefaultGroup.PERSONAL_WORKSPACE.value:
            group_id = DefaultGroup.LENS_GROUP_ID_PERSONAL_WORKSPACE.value
        elif group == DefaultGroup.ALL_MEMBERS.value:
            group_id = DefaultGroup.LENS_GROUP_ID_ALL_MEMBERS.value
        else:
            group_id = self.groups[group]

        studies_info = connector.get_all_studies_info_in_group(group_id=group_id, species=species)
        studies_info = [
            {
                **info,
                ConnectorKeys.SPECIES.value: species,
                ConnectorKeys.GROUP_ID.value: self.groups[group],
            }
            for info in studies_info
        ]
        return studies_info

    def convert_data_from_lens(self, study_info: Union[dict, List[dict]]):
        if isinstance(study_info, dict):
            study_info = [study_info]

        for info in study_info:
            self.openapi.convert_from_lens(
                f"ST-{info['id']}",
                info["title"],
                info[ConnectorKeys.GROUP_ID.value],
                info[ConnectorKeys.SPECIES.value],
                info["id"],
            )



from typing import Optional, Literal

from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData
from ... import _constants as constants


def pca(
    adata: ConnectorAnnData,
    normalize_method: Optional[
        Literal[
            constants.NormalizeMethod.RAW,
            constants.NormalizeMethod.LOG1P_NORMALIZE,
            constants.NormalizeMethod.SQRT_NORMALIZE,
        ]
    ] = constants.NormalizeMethod.LOG1P_NORMALIZE,
    title: Optional[str] = None,
    batch_key: Optional[str] = None,
    n_top_genes: Optional[int] = None,
    **kwargs,
):
    """
    Principal Component Analysis

    Parameters
    ----------
    normalize_method: `Literal["raw", "log1p-normalized", "sqrt-normalized"]`, Default: "log1p-normalized"
        Using raw expression or scaling (log1p, square root) to normalize when running PCA.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    batch_key: `Optional[str]`, Default: None
        Metadata which defines batch effect.
        If provided, using Harmony to remove batch effect.
    n_top_genes: `Optional[int]`, Default: None
        Number of top highly variable genes.
        If provided, running highly variable genes to select genes for PCA.
    """
    analysis = adata.self_init(Analysis)
    analysis.embeddings.pca(
        normalize_method=normalize_method,
        title=title,
        batch_key=batch_key,
        n_top_genes=n_top_genes,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def scvi(
    adata: ConnectorAnnData,
    n_latents: int = 20,
    title: Optional[str] = None,
    batch_key: Optional[str] = None,
    n_top_genes: Optional[int] = 2000,
    **kwargs,
):
    """
    Reducing dimensions with scVI

    Parameters
    ----------
    n_latents: `int`, Default: 20
        Number of latents.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    batch_key: `Optional[str]`, Default: None
        Metadata which defines batch effect.
        If provided, using Harmony to remove batch effect.
    n_top_genes: `Optional[int]`, Default: 2000
        Number of top highly variable genes.
        If provided, running highly variable genes to select genes for PCA.
    """
    analysis = adata.self_init(Analysis)
    analysis.embeddings.scvi(
        n_latents=n_latents,
        title=title,
        batch_key=batch_key,
        n_top_genes=n_top_genes,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()



import re
from pathlib import Path
from setuptools import setup, find_packages


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

__version__ = (this_directory / "spatialx_connector/_version.py").read_text().splitlines()[0]
__version__ = re.search(r'"(\d+\.\d+\.\d+)"', __version__).group(1)

BASE_PACKAGE = "spatialx_connector"
BASE_IMPORT = "spatialx_connector"


def _install_requires():
    return [
        "pydantic==2.7.2",
        "requests==2.32.3",
        "tqdm==4.66.4",
        "zarr== 2.18.2",
        "scanpy==1.11.0",
        "matplotlib==3.10.0",
        "spatialdata==0.2.5",
        "anndata==0.10.8",
        "dask==2024.8.2",
        "dask-expr==1.1.13",
    ]


setup(
    name=BASE_PACKAGE,
    version=__version__,
    author="BioTuring",
    author_email="support@bioturing.com",
    url="https://alpha.bioturing.com",
    description="BioTuring SpatialX Connector",
    package_dir={BASE_IMPORT: "spatialx_connector"},
    packages=[BASE_IMPORT, *find_packages()],
    zip_safe=False,
    install_requires=_install_requires(),
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from typing import Optional, Literal

from ..._api import PyAPI
from ..._analysis import Analysis
from ..._anndata import ConnectorAnnData
from ..._constants import Species


def metadata_reference(
    adata: ConnectorAnnData,
    cluster_key: str,
    species: Literal[Species.HUMAN, Species.MOUSE],
    annotation_type: Literal["sub", "major"] = "sub",
    title: Optional[str] = None,
    **kwargs,
):
    """
    Spatial Region Segmentation.

    Parameters
    ----------
    cluster_key: `str`
        Cluster to predict cell types.
    species: `str`
        Species of data.
    annotation_type: `Literal["sub", "major"]`, Default: "sub"
        Type of annotation: Major or sub cell types.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    """
    analysis = adata.self_init(Analysis)
    analysis.prediction.metadata_reference(
        cluster_key=cluster_key,
        species=species,
        anno_type=annotation_type,
        title=title,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()


def list_deconvolution_models(adata: ConnectorAnnData):
    return adata.self_init(PyAPI).list_deconvolution_models()


def deconvolution(
    adata: ConnectorAnnData,
    model_id: str,
    title: Optional[str] = None,
    n_cells_per_location: int = 30,
    detection_alpha: int = 20,
    **kwargs,
):
    """
    Spatial Region Segmentation.

    Parameters
    ----------
    model_id: `str`
        Deconvolution model id.
    title: `Optional[str]`, Default: None
        Title of analysis, it will be used as name of embeddings.
        If not provided, a default name will be used.
    n_cells_per_location:  `int`
        Estimation of absolute cell abundance is guided using informed prior on the number of cells
    """
    analysis = adata.self_init(Analysis)
    analysis.prediction.deconvolution(
        model_id=model_id,
        title=title,
        n_cells_per_location=n_cells_per_location,
        detection_alpha=detection_alpha,
        info_args=adata._extend_information,
        **kwargs,
    )
    adata.update()



from typing import Union, Dict, List
from functools import cached_property
import logging

import numpy as np
import pandas as pd

from ._api import Connector, OpenAPI
from ._spatialdata import SpatialData
from ._utils import left_aligning, right_aligning, confirm
from ._constants import ConnectorKeys, EnableStatus, SpatialAttrs


class Data(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        data: Union[dict, str] = {},
        **kwargs,
    ):
        super().__init__(domain, token, **kwargs)
        self.__openapi: OpenAPI = self.self_init(OpenAPI)

        if isinstance(data, str):
            data = self.__openapi.get_sample_data_detail(data_id=data)
        self.__details = data

        self.data_id = self.__details[ConnectorKeys.DATA_ID.value]
        self.sample_id = self.__details[ConnectorKeys.SAMPLE_ID.value]
        self.study_id = self.__details[ConnectorKeys.STUDY_ID.value]
        self.title = self.__details[ConnectorKeys.TITLE.value]
        self.technology = self.__details[ConnectorKeys.TECHNOLOGY.value]
        self.enable_status = self.__details[ConnectorKeys.ENABLE_STATUS.value]

    def update(self):
        self = self.self_init(Data, data=self.data_id)

    def repr(self):
        return f"[Data - {self.data_id}] {self.title} "\
            f"(Technology: {self.technology} - Status: {EnableStatus.get_status(self.enable_status)})"

    def __repr__(self):
        text = self.repr() + "\n with elements:"

        mapping_elements = {
            "images": "Images",
            "segmentation": "Segmentation",
            "spot": "Spots",
            "cell_centers": "Cells centers",
            "transcripts": "Transcripts",
            "annotated_data": "AnnData",
        }

        for k in mapping_elements.keys():
            if k not in self.elements:
                continue
            values = self.elements[k]
            text += f"\n\t{mapping_elements[k]}: " + ", ".join([self.id2name[v] for v in values])
        return text

    def rename(self, new_name: str):
        if self.title == new_name:
            return
        self.__openapi.rename_sample_data(self.data_id, new_name)
        self.update()

    def delete(self):
        logger = logging.getLogger("spatialx_sdks_stdout")
        logger.info(f"Do you confirm to delete data {self.repr()}?")

        text = f"Do you confirm to delete data {self.repr()}?"
        if not confirm(text):
            logger.info("Canceled!")
            return
        res = self.__openapi.delete_sample_data(self.data_id)
        self.update()
        return res

    @cached_property
    def elements(self) -> Dict[str, List[str]]:
        results: dict = self.__details.get("map_submit_result", {})

        elements: Dict[str, List[str]] = {}
        for value in results.values():
            if not isinstance(value, dict):
                continue
            for k, v in value.items():
                if k not in elements:
                    elements[k] = []
                elements[k].extend(v.keys() if isinstance(v, dict) else [v])
        return elements

    @cached_property
    def id2name(self) -> dict:
        id2name = {}
        spdata = self.self_init(SpatialData, data_id=self.data_id)
        for k, values in self.elements.items():
            k = SpatialAttrs.SPATIAL_ELEMENT_MAPPING.value[k]
            for v in values:
                id2name[v] = getattr(spdata, k).get_name_by_id(v)
        return id2name

    @cached_property
    def name2id(self) -> dict:
        names, counts = np.unique(list(self.id2name.values))
        names = names[counts == 1]
        return {v: k for k, v in self.id2name.items() if v in names}

    def __name2id(self, element_id: str):
        if element_id not in self.elements:
            if element_id not in self.name2id:
                raise ValueError(f"Element {element_id} is not found or is duplicated in data.")
            element_id = self.name2id[element_id]
        return element_id

    def delete_element(self, element_id: str):
        logger = logging.getLogger("spatialx_sdks_stdout")
        element_id = self.__name2id(element_id)
        logger.info(f"Do you confirm to delete element {self.id2name[element_id]}?")

        text = f"Do you confirm to delete element {self.id2name[element_id]}?"
        if not confirm(text):
            logger.info("Canceled!")
            return
        res = self.__openapi.delete_sample_data_element(self.data_id, element_id)
        self.update()
        return res

    def add_elements(
        self,
        title: str,
        adding_types: List[str],
        paths: Dict[str, str] = {},
        args: Dict[str, str] = {},
        kwargs: Dict[str, str] = {},
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        title: ``str``
            Title of elements
        adding_types: `List[str]`
            Type of element to adding, defined in [
                `spatialx_connector.ImagesSubmission`,
                `spatialx_connector.SegmentationSubmission`,
                `spatialx_connector.TrasncriptsSubmission`,
                `spatialx_connector.ExpressionSubmission`,
            ]
        paths: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and their paths.
        args: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and values.

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        res = self.__openapi.add_sample_data_element(
            title,
            self.study_id,
            self.sample_id,
            self.data_id,
            identities=adding_types,
            files=[
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in paths.items()
            ],
            args=[
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in args.items()
            ] + [
                {
                    # Ignore default types of technologies.
                    ConnectorKeys.KEY.value: "ignore_technology_elements",
                    ConnectorKeys.VALUE.value: True,
                }
            ],
            kwargs=[
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in kwargs.items()
            ]
        )
        return res

    @cached_property
    def spatialdata(self) -> SpatialData:
        return self.self_init(SpatialData, data_id=self.data_id)


class Sample(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        data: Union[dict, str] = {},
        **kwargs,
    ):
        super().__init__(domain, token, **kwargs)
        self.__openapi: OpenAPI = self.self_init(OpenAPI)

        if isinstance(data, dict) and ("sample" not in data or "sample_data" not in data):
            data = data[ConnectorKeys.SAMPLE_ID.value]

        if isinstance(data, str):
            data = self.__openapi.get_sample_detail(sample_id=data)
        self.__sampe_details = data["sample"]
        self.__data_details = data["sample_data"]

        self.sample_id = self.__sampe_details[ConnectorKeys.SAMPLE_ID.value]
        self.study_id = self.__sampe_details[ConnectorKeys.STUDY_ID.value]
        self.title = self.__sampe_details[ConnectorKeys.TITLE.value]
        self.enable_status = self.__sampe_details[ConnectorKeys.ENABLE_STATUS.value]

        self.sample_data: Dict[str, Data] = {}
        self.index_sample_data = {}
        for i, data in enumerate(self.__data_details):
            if data[ConnectorKeys.ENABLE_STATUS.value] >= 3:
                continue
            self.sample_data[data[ConnectorKeys.DATA_ID.value]] = self.self_init(Data, data=data)
            self.index_sample_data[data[ConnectorKeys.DATA_ID.value]] = i
            self.index_sample_data[i] = data[ConnectorKeys.DATA_ID.value]

    def update(self):
        self = self.self_init(Sample, data=self.sample_id)

    def __repr__(self):
        text = f"[Sample - {self.sample_id}] {self.title} (Status: {EnableStatus.get_status(self.enable_status)})"
        for data in self.sample_data.values():
            text += f"\n\t{self.index_sample_data[data.data_id]}. " + data.repr()
        return text

    def __getitem__(self, key) -> Data:
        if isinstance(key, int):
            key = self.index_sample_data[key]
        return self.sample_data[key]

    def __iter__(self):
        for sample_data in self.sample_data.values():
            yield sample_data

    def rename(self, new_name: str):
        if self.title == new_name:
            return
        self.__openapi.rename_sample(self.sample_id, new_name)
        self.update()

    def delete(self):
        logger = logging.getLogger("spatialx_sdks_stdout")
        desc = "\t" + self.__repr__().replace('\n', '\n\t')
        logger.info(f"Do you confirm to delete sample and all its data below:\n{desc}")

        text = f"Delete sample `{self.title}` with its {len(self.sample_data)} data?"
        if not confirm(text):
            logger.info("Canceled!")
            return
        res = self.__openapi.delete_sample(self.sample_id)
        self.update()
        return res


class Study(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        data: Union[dict, str] = {},
        **kwargs,
    ):
        super().__init__(domain, token, **kwargs)
        self.__openapi: OpenAPI = self.self_init(OpenAPI)

        if isinstance(data, str):
            data = self.__openapi.get_study_detail(study_id=data)
        self.__details = data

        self.study_id = self.__details[ConnectorKeys.STUDY.value][ConnectorKeys.STUDY_ID.value]
        self.title = self.__details[ConnectorKeys.STUDY.value][ConnectorKeys.TITLE.value]
        self.species = self.__details[ConnectorKeys.STUDY.value][ConnectorKeys.SPECIES.value]

        self.sample: Dict[str, Sample] = {}
        self.index_sample = {}
        for i, sample in enumerate(self.__details["samples"]):
            if sample[ConnectorKeys.ENABLE_STATUS.value] >= 3:
                continue
            self.sample[sample[ConnectorKeys.SAMPLE_ID.value]] = self.self_init(Sample, data=sample)
            self.index_sample[sample[ConnectorKeys.SAMPLE_ID.value]] = i
            self.index_sample[i] = sample[ConnectorKeys.SAMPLE_ID.value]

    def update(self):
        self = self.self_init(Study, data=self.study_id)

    def __repr__(self):
        text = f"{self.title} (Species: {self.species})"
        for sample in self.sample.values():
            text += f"\n\t{self.index_sample[sample.sample_id]}. " + sample.__repr__().replace("\n", "\n\t")
        return text

    def __getitem__(self, key) -> Sample:
        if isinstance(key, int):
            key = self.index_sample[key]
        return self.sample[key]

    def __iter__(self):
        for sample in self.sample.values():
            yield sample

    def rename(self, new_name: str):
        if self.title == new_name:
            return
        self.__openapi.rename_study(self.study_id, new_name)
        self.update()

    def delete(self):
        logger = logging.getLogger("spatialx_sdks_stdout")
        logger.info(f"Do you confirm to delete study {self.title}?")

        text = f"Do you confirm to delete study {self.title}?"
        if not confirm(text):
            logger.info("Canceled!")
            return
        res = self.__openapi.delete_study(self.study_id)
        self.update()
        return res


class Studies(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        data: Union[list, dict] = {},
        **kwargs,
    ):
        super().__init__(domain, token, **kwargs)
        self.__openapi: OpenAPI = self.self_init(OpenAPI)

        if isinstance(data, dict):
            data: list = self.__openapi.list_study(**data)["list"]
        self.__details = data

        fields = [
            ConnectorKeys.STUDY_ID.value,
            ConnectorKeys.TITLE.value,
            ConnectorKeys.TOTAL_SAMPLE.value,
            ConnectorKeys.TOTAL_DATA.value,
        ]
        info: Dict[str, List] = {"ID": []}
        info.update({field: [] for field in fields})
        for i, study in enumerate(self.__details):
            info["ID"].append(i)
            for field in fields:
                info[field].append(study[field])
        study_id = info[ConnectorKeys.STUDY_ID.value]
        self.info: pd.DataFrame = pd.DataFrame(info, index=study_id)

    def __repr__(self):
        columns = ["", "Study ID", "Title", "Total Sample", "Total Data"]
        column_aligning = [left_aligning, left_aligning, left_aligning, right_aligning, right_aligning]

        sizes = [
            max(
                len(columns[i]),
                np.max([len(str(v)) for v in self.info.values[:, i]])
            )
            for i in range(len(column_aligning))
        ]
        text = " │ ".join([
            left_aligning(column, sizes[i])
            for i, column in enumerate(columns)
        ]) + " │"
        text += "\n" + "─┼─".join(["─" * size for size in sizes]) + "─┤"
        for values in self.info.values:
            text += "\n" + " │ ".join([
                column_aligning[i](values[i], sizes[i])
                for i in range(len(values))
            ]) + " │"
        return text

    def head(self, n: int):
        return self.info.head(n)

    def tail(self, n: int):
        return self.info.tail(n)

    def __getitem__(self, k: Union[int, str]) -> Study:
        if isinstance(k, int):
            k = str(self.info.index.values[k])
        return self.self_init(Study, data=k)



import sys
import time
import logging
import functools
from enum import Enum
from typing import Optional

import requests
from requests.adapters import Retry
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth

from ._constants import ConnectorKeys


TAB_STRING: str = " " * 4


DEFAULT_LOGGER = logging.getLogger("spatialx_sdks_stdout")
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter(
    fmt="[%(asctime)s] [%(process)d] [%(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
))
DEFAULT_LOGGER.addHandler(handler)
DEFAULT_LOGGER.setLevel(logging.INFO)


def time_cache(max_age, maxsize=128, typed=False):
    """Least-recently-used cache decorator with time-based cache invalidation.

    Args:
        max_age: Time to live for cached results (in seconds).
        maxsize: Maximum cache size (see `functools.lru_cache`).
        typed: Cache on distinct input types (see `functools.lru_cache`).
    """
    def _decorator(fn):
        @functools.lru_cache(maxsize=maxsize, typed=typed)
        def _new(*args, __time_salt, **kwargs):
            return fn(*args, **kwargs)

        @functools.wraps(fn)
        def _wrapped(*args, **kwargs):
            return _new(*args, **kwargs, __time_salt=int(time.time() / max_age))

        return _wrapped

    return _decorator


class RequestMode:
    POST: str = "post"
    PUT: str = "put"
    GET: str = "get"


def create_requests_session():
    session = requests.Session()
    retries = Retry(
        total=5,
        backoff_factor=0.1
    )
    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.mount('https://', HTTPAdapter(max_retries=retries))
    return session


def parse_enum_to_value(data):
    if isinstance(data, list) or isinstance(data, tuple):
        return parse_list_enum_to_value(data)
    if isinstance(data, dict):
        return parse_dict_enum_to_value(data)
    return data.value if isinstance(data, Enum) else data


def parse_list_enum_to_value(data: list):
    return [parse_enum_to_value(v) for v in data]


def parse_dict_enum_to_value(data: dict):
    return {parse_enum_to_value(k): parse_enum_to_value(v) for k, v in data.items()}


def request(
    url: str,
    params={},
    data={},
    json={},
    headers={},
    files={},
    mode: str = RequestMode.POST,
    verify: bool = False,
    authentication: Optional[HTTPBasicAuth] = None,
):
    session = create_requests_session()
    r: requests.Response = getattr(session, mode)(
        url,
        params=parse_enum_to_value(params),
        json=parse_enum_to_value(json),
        data=parse_enum_to_value(data),
        headers=parse_enum_to_value(headers),
        files=parse_enum_to_value(files),
        verify=verify,
        timeout=1800,
        auth=authentication,
    )

    if r.status_code >= 400:
        raise Exception(
            f"Call request to {url} failed with status code {r.status_code}, response {r.text}"
        )
    try:
        response = r.json()
    except:
        response = r.text
    r.close()
    return response


def parse_to_str(data, tab_level: int = 0) -> str:
    if isinstance(data, list):
        return list_to_str(data, tab_level)
    if isinstance(data, tuple):
        return tuple_to_str(data, tab_level)
    if isinstance(data, dict):
        return dict_to_str(data, tab_level)
    return f"{data}"


def list_to_str(data: list, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "[""]"
    return "[\n" + "".join([
        TAB_STRING * (tab_level + 1) +
        f"{parse_to_str(value, tab_level + 1)}\n" for value in data
    ]) + TAB_STRING * tab_level + "]"


def tuple_to_str(data: list, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "("")"
    return "(\n" + "".join([
        TAB_STRING * (tab_level + 1) +
        f"{parse_to_str(value, tab_level + 1)}\n" for value in data
    ]) + TAB_STRING * tab_level + ")"


def dict_to_str(data: dict, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "{""}"
    return "{\n" + "".join([
        f"{TAB_STRING * (tab_level + 1)}{key}: "
        f"{parse_to_str(value, tab_level + 1)}\n"
        for key, value in data.items()
    ]) + TAB_STRING * tab_level + "}"


def format_print(data):
    print(parse_to_str(data))


def get_chunk_size(chunk_size: int, file_size: int) -> int:
    if chunk_size > 0:
        if chunk_size > ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value:
            chunk_size = ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value
        return chunk_size

    if file_size < 15*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_SMALL_SIZE.value
    elif file_size < 100*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_MEDIUM_SIZE.value
    elif file_size < 1024*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_NORMAL_SIZE.value

    return ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value


def right_aligning(value: str, length: int, char: str = " "):
    value = str(value)
    length = max(length, len(value))
    return char * (length - len(value)) + value


def left_aligning(value: str, length: int, char: str = " "):
    value = str(value)
    length = max(length, len(value))
    return value + char * (length - len(value))


def confirm(text: str):
    text += "\nPlease type `Yes` to confirm this action: "
    return input(text).lower() == "yes"



import os
import time
import base64
import logging
import traceback
from functools import cached_property
from urllib.parse import urlparse, parse_qs
from typing import Union, List, Dict, Optional


from pydantic import BaseModel  # pylint: disable=no-name-in-module
from requests.auth import HTTPBasicAuth

import zarr
import zarr.storage
import pandas as pd

from . import _constants as constants
from ._utils import RequestMode
from ._utils import request
from ._utils import parse_to_str


DEBUG_MODE = os.environ.get("DEBUG_MODE", "") == "TRUE"
DEFAULT_LOGGER = logging.getLogger("spatialx_sdks_stdout")


class Connector:
    def __init__(
        self,
        domain: str,
        token: str,
        verify_ssl: bool = False,
        authentication: Optional[HTTPBasicAuth] = None,
        **_,
    ):
        self._token = token
        self._verify_ssl = verify_ssl
        self._authentication = authentication

        headers = {constants.TOKEN_KEY: self._token}
        if self._authentication is not None:
            if isinstance(self._authentication, HTTPBasicAuth):
                authentication = (self._authentication.username, self._authentication.password)
            else:
                authentication = self._authentication
            token = base64.b64encode(f"{authentication[0]}:{authentication[1]}".encode("utf-8")).decode("utf-8")
            headers["Authorization"] = f"Basic {token}"
        self._headers = headers

        link = urlparse(domain)
        if len(link.netloc) == 0:
            raise Exception("invalid domain: {}".format(domain))

        params = dict(parse_qs(link.query))
        params = {k: v[0] for k, v in params.items()}
        self.params = params
        self._domain = "{}://{}{}".format(link.scheme, link.netloc, link.path)

    def self_init(self, obj_type, *args, **kwargs):
        return obj_type(
            self._domain, self._token, *args,
            verify_ssl=self._verify_ssl,
            authentication=self._authentication,
            **kwargs,
        )

    def request(
            self,
            url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ):
        for k, v in self.params.items():
            params[k] = v

        res: Union[str, dict] = request(
            url=url.replace(os.sep, "/"),
            params=params,
            json=json,
            data=data,
            files=files,
            headers=self._headers,
            mode=mode,
            verify=self._verify_ssl,
            authentication=self._authentication,
        )
        if isinstance(res, str):
            return res

        if res.get(constants.ConnectorKeys.STATUS.value, 0) != 0:
            raise Exception(parse_to_str(res))

        return res.get(constants.ConnectorKeys.MESSAGE.value, res)

    def post_pyapi_request(
            self, url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ) -> Union[BaseModel, str]:
        return self.request(
            os.path.join(self._domain, constants.V1_API, url),
            params=params, json=json, data=data,
            mode=mode, files=files,
        )

    def post_openapi_request(
            self, url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ) -> dict:
        return self.request(
            os.path.join(self._domain, constants.OPEN_API, url),
            params=params, json=json, data=data,
            mode=mode, files=files,
        )

    def _get_zarr_path(self, path: str):
        return os.path.join(self._domain, constants.UCC, path)

    def _open_zarr_root(self, path: str):
        url = self._get_zarr_path(path)
        return zarr.storage.FSStore(url=url, client_kwargs={"headers": self._headers})

    def tracking_log(self, log_path: str, max_time: int = 60):
        log_path = os.path.join(self._domain, constants.UCC, log_path)
        curr_row = 0
        curr_time = time.time()
        for _ in range(max_time):
            if time.time() - curr_time > max_time:
                break
            data = self.request(log_path, mode=RequestMode.GET)
            if data.endswith("\n"):
                data = data[:-1]
            lines = data.split("\n")
            if len(lines) > curr_row:
                print("\n".join(lines[curr_row:]))
            curr_row = len(lines)
            if "DONE." in data[-1000:] or "maximum retry limit" in data[-1000:]:
                break
            time.sleep(1)


class PyAPI(Connector):
    def parse_data_information(self, name: str, technology: str, data_path: str) -> dict:
        return self.post_pyapi_request(
            url=constants.PARSE_DATA_URL,
            json={
                constants.ConnectorKeys.DATA_NAME.value: name,
                constants.ConnectorKeys.TECHNOLOGY.value: technology,
                constants.ConnectorKeys.DATA_PATH.value: data_path,
            }
        )

    def rename_image_channels(
        self,
        study_id: str,
        sample_id: str,
        images_id: str,
        channel_names: List[str],
    ) -> dict:
        return self.post_pyapi_request(
            url=constants.RENAME_IMAGE_CHANNELS,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.IMAGES_ID.value: images_id,
                constants.ConnectorKeys.CHANNEL_NAMES.value: channel_names,
            }
        )

    def simplify_segmentation(
        self,
        study_id: str,
        sample_id: str,
        segmentation_id: str,
        tolerance: Optional[float],
    ) -> dict:
        return self.post_pyapi_request(
            url=constants.SIMPLIFY_SEGMENTATION,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.SEGMENTATION_ID.value: segmentation_id,
                constants.ConnectorKeys.TOLERANCE.value: tolerance,
            }
        )

    def import_metadata(
        self,
        study_id: str,
        sample_id: str,
        table_id: str,
        csv_path: str,
    ) -> dict:
        return self.post_pyapi_request(
            url=constants.IMPORT_METADATA,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.TABLE_ID.value: table_id,
                constants.ConnectorKeys.CSV_PATH.value: csv_path,
            }
        )

    def list_deconvolution_models(self) -> dict:
        data = self.post_pyapi_request(
            url=constants.LIST_DECONVOLUTION_MODELS,
            mode=RequestMode.GET,
        )
        return pd.DataFrame(data).T


class OpenAPI(Connector):
    @property
    def info(self):
        return self.post_openapi_request(
            url=constants.INFO_URL,
            mode=RequestMode.GET,
        )

    @property
    def mounts(self):
        return self.post_openapi_request(
            url=constants.EXTERNAL_MOUNT_URL,
            mode=RequestMode.GET,
        )

    def list_s3(self, offset: int = 0, limit: int = 100):
        return self.post_openapi_request(
            url=constants.LIST_S3,
            mode=RequestMode.POST,
            data={
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            }
        )

    @property
    def s3(self):
        return self.list_s3()

    @property
    def groups(self):
        return self.post_openapi_request(
            url=constants.GROUPS_URL,
            mode=RequestMode.GET,
        )

    def list_dir(self, path: str, ignore_hidden: bool = True):
        return self.post_openapi_request(
            constants.LIST_URL,
            data={
                constants.ConnectorKeys.PATH.value: path,
                constants.ConnectorKeys.IGNORE_HIDDEN.value: ignore_hidden,
            }
        )

    def create_study(
        self,
        group_id: str,
        species: str,
        title: str,
        create_type: int = constants.StudyType.SPATIAL_STUDY_TYPE_NUMBER.value,
    ):
        return self.post_openapi_request(
            url=constants.CREATE_STUDY_URL,
            json={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.TITLE.value: title,
                constants.ConnectorKeys.TYPE.value: create_type,
            }
        )

    def list_study(
        self,
        group_id: str,
        species: str,
        limit: int = 50,
        offset: int = 0,
        active: int = constants.StudyStatus.PROCESSING_STATUS.value,
        compare: int = constants.StudyFilter.NOT_LARGER.value,
    ):
        return self.post_openapi_request(
            url=constants.LIST_STUDY_URL,
            data={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.ACTIVE.value: active,
                constants.ConnectorKeys.COMPARE.value: compare,
            }
        )

    def get_study_detail(self, study_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_STUDY_URL,
            params={
                constants.ConnectorKeys.KEY.value: study_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def update_study(
        self,
        study_id: str,
        title,
        description,
        email_id,
        group_id,
        species,
        publication: List[str] = [],
        tags: List[str] = [],
        author: List[str] = [],
    ):
        return self.post_openapi_request(
            url=constants.UPDATE_STUDY_URL,
            params={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.TITLE.value: title,
                constants.ConnectorKeys.DESCRIPTION.value: description,
                constants.ConnectorKeys.EMAIL_ID.value: email_id,
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.PUBLICATION.value: publication,
                constants.ConnectorKeys.TAGS.value: tags,
                constants.ConnectorKeys.AUTHOR.value: author,
            },
            mode=RequestMode.POST,
        )

    def rename_study(
        self,
        study_id: str,
        title: str,
    ):
        details = self.get_study_detail(study_id)[constants.ConnectorKeys.STUDY.value]
        return self.update_study(
            study_id=details[constants.ConnectorKeys.STUDY_ID.value],
            title=title,
            description=details[constants.ConnectorKeys.DESCRIPTION.value],
            email_id=details[constants.ConnectorKeys.EMAIL_ID.value],
            group_id=details[constants.ConnectorKeys.GROUP_ID.value],
            species=details[constants.ConnectorKeys.SPECIES.value],
            publication=details[constants.ConnectorKeys.PUBLICATION.value],
            tags=details[constants.ConnectorKeys.TAGS.value],
            author=details[constants.ConnectorKeys.AUTHOR.value],
        )

    def delete_study(self, study_id: str):
        return self.post_openapi_request(
            url=constants.DELETE_STUDY_URL,
            params={
                constants.ConnectorKeys.KEY.value: study_id,
                constants.ConnectorKeys.CLEANUP.value: False,
            },
            mode=RequestMode.POST,
        )

    def create_sample(
        self,
        study_id: str,
        name: str,
        data: List[dict],
    ):
        return self.post_openapi_request(
            url=constants.CREATE_SAMPLE_URL,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.NAME.value: name,
                constants.ConnectorKeys.DATA.value: data,
            }
        )

    def add_sample_data(
        self,
        study_id: str,
        sample_id: str,
        data: List[dict],
    ):
        return self.post_openapi_request(
            url=constants.ADD_SAMPLE_DATA_URL,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.DATA.value: data,
            }
        )

    def add_sample_data_element(
        self,
        title: str,
        study_id: str,
        sample_id: str,
        data_id: str,
        identities: List[str] = [],
        files: List[Dict[str, str]] = [],
        folders: List[Dict[str, str]] = [],
        args: List[Dict[str, str]] = [],
        kwargs: List[Dict[str, str]] = [],
    ):
        return self.post_openapi_request(
            url=constants.ADD_URL,
            json={
                constants.ConnectorKeys.TITLE.value: title,
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.DATA_ID.value: data_id,
                constants.ConnectorKeys.FILES.value: files,
                constants.ConnectorKeys.FOLDERS.value: folders,
                constants.ConnectorKeys.ARGS.value: args,
                constants.ConnectorKeys.KWARGS.value: kwargs,
                constants.ConnectorKeys.IDENTITIES.value: identities,
            },
        )

    def list_sample(
        self,
        study_id: str,
        limit: int = 50,
        offset: int = 0,
        need_data: bool = False,
    ):
        return self.post_openapi_request(
            url=constants.LIST_SAMPLE_URL,
            params={
                constants.ConnectorKeys.KEY.value: study_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.NEED_DATA.value: need_data,
            },
            mode=RequestMode.GET,
        )

    def get_sample_detail(self, sample_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_SAMPLE_URL,
            params={
                constants.ConnectorKeys.KEY.value: sample_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def rename_sample(self, sample_id: str, new_name: str):
        return self.post_openapi_request(
            url=constants.RENAME_SAMPLE_URL,
            json={
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.NEW_NAME.value: new_name,
            }
        )

    def delete_sample(self, sample_id: str):
        return self.post_openapi_request(
            url=constants.DELETE_SAMPLE_URL,
            json={constants.ConnectorKeys.SAMPLE_ID.value: sample_id}
        )

    def get_sample_data_detail(self, data_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_SAMPLE_DATA_URL,
            params={
                constants.ConnectorKeys.KEY.value: data_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def rename_sample_data(self, data_id: str, new_name: str):
        return self.post_openapi_request(
            url=constants.RENAME_SAMPLE_DATA_URL,
            json={
                constants.ConnectorKeys.DATA_ID.value: data_id,
                constants.ConnectorKeys.NEW_NAME.value: new_name,
            }
        )

    def delete_sample_data(self, data_id: str):
        return self.post_openapi_request(
            url=constants.DELETE_SAMPLE_DATA_URL,
            json={constants.ConnectorKeys.DATA_ID.value: data_id}
        )

    def delete_sample_data_element(self, data_id: str, element_id: str):
        return self.post_openapi_request(
            url=constants.DELETE_SAMPLE_DATA_ELEMENT_URL,
            json={
                constants.ConnectorKeys.DATA_ID.value: data_id,
                constants.ConnectorKeys.ELEMENT_ID.value: element_id,
            }
        )

    def list_public_study(
        self,
        group_id: str,
        species: str,
        limit: int = 50,
        offset: int = 0,
        active: int = constants.StudyStatus.PROCESSING_STATUS.value,
    ):
        return self.post_openapi_request(
            url=constants.LIST_PUBLIC_STUDY_URL,
            json={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.ACTIVE.value: active,
            }
        )

    def upload_file(
        self, file_path: str,
        folder_name: str,
        upload_id: str,
        is_chunk: bool = False,
    ):
        file = open(file_path, "rb")
        resp = self.post_openapi_request(
            url=constants.UPLOAD_FILE_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_IS_CHUNK.value: is_chunk,
            },
            files={
                constants.ConnectorKeys.UPLOAD_FILE_DATA.value: file,
            },
        )
        file.close()
        return resp

    def upload_chunk_start(self, folder_name: str, parent_is_file: int):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_START_URL,
            json={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
            }
        )

    def upload_chunk_process(
        self,
        chunk_size: int,
        file_size: int,
        offset: int,
        file_name: str,
        folder_name: str,
        upload_id: str,
        path: str,
        sending_index: int,
        parent_is_file: int,
        file_data: list[str],
    ):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_PROCESS_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
                constants.ConnectorKeys.UPLOAD_CHUNK_SIZE.value: chunk_size,
                constants.ConnectorKeys.UPLOAD_FILE_SIZE.value: file_size,
                constants.ConnectorKeys.UPLOAD_OFFSET.value: offset,
                constants.ConnectorKeys.UPLOAD_FILE_NAME.value: file_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_PATH.value: path,
                constants.ConnectorKeys.UPLOAD_SENDING_INDEX.value: sending_index,
            },
            files={
                constants.ConnectorKeys.UPLOAD_FILE_DATA.value: file_data,
            }
        )

    def upload_chunk_merge(
        self,
        total_chunk: int,
        file_name: str,
        folder_name: str,
        upload_id: str,
        path: str,
        parent_is_file: int,
        move_to_parent: bool,
    ):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_MERGE_URL,
            json={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
                constants.ConnectorKeys.UPLOAD_TOTAL_CHUNK.value: total_chunk,
                constants.ConnectorKeys.UPLOAD_FILE_NAME.value: file_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_PATH.value: path,
                constants.ConnectorKeys.UPLOAD_MOVE_TO_PARENT.value: move_to_parent,
            }
        )

    def upload_folder_finish(self, folder_name: str, upload_id: str):
        return self.post_openapi_request(
            url=constants.UPLOAD_FOLDER_FINISH_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
            },
        )

    def convert_from_lens(
        self, study_id: str, study_name: str,
        group_id: str, species: str, lens_data_path: str
    ):
        return self.post_openapi_request(
            url=constants.STUDY_CONVERT_FROM_LENS_URL,
            data={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.TITLE.value: study_name,
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LENS_DATA_PATH.value: lens_data_path,
            }
        )


class ZarrElement(Connector):
    def __init__(
        self,
        domain: str,
        token: str,
        path: str = "",
        verify_ssl: bool = False,
        authentication: Optional[HTTPBasicAuth] = None,
        elements: List[str] = None,
        extend_information: dict = None,
    ):
        super().__init__(domain, token, verify_ssl=verify_ssl, authentication=authentication)
        self._path = path
        self._elements = elements
        self._extend_information = extend_information if extend_information is not None else {}

    def self_init(self, obj_type, path=None, elements=None, extend_information=None):
        information = self._extend_information
        if extend_information is not None:
            information.update(extend_information)

        if path is None:
            path = self._path

        return obj_type(
            self._domain,
            self._token,
            verify_ssl=self._verify_ssl,
            authentication=self._authentication,
            path=path,
            elements=elements,
            extend_information=information,
        )

    def update(self):
        return self.__init__(
            self._domain,
            self._token,
            verify_ssl=self._verify_ssl,
            authentication=self._authentication,
            path=self._path,
            elements=self._elements,
            extend_information=self._extend_information,
        )

    @cached_property
    def group(self) -> zarr.Group:
        return zarr.Group(self._open_zarr_root(self._path), read_only=True)

    @property
    def attrs(self):
        try:
            res = self.group.attrs.asdict()
        except Exception as e:
            import logging

            logger = logging.getLogger("spatialx_sdks_stdout")
            if DEBUG_MODE:
                logger.error("".join(traceback.format_exception(e)))
            else:
                logger.warning(e)
            res = {}

        return res

    @property
    def _name2id(self) -> Dict[str, str]:
        elements = self.attrs.get("elements", {})
        if self._elements is None:
            return elements
        return {name: ele_id for name, ele_id in elements.items() if ele_id in self._elements}

    def get_id_by_name(self, name: Union[str, List[str]], name2id=None) -> str:
        if name2id is None:
            name2id = self._name2id

        if isinstance(name, str):
            return name2id.get(name, name)
        return [self.get_id_by_name(n, name2id) for n in name]

    def get_name_by_id(self, elem_id: Union[str, List[str]], id2name=None) -> str:
        if id2name is None:
            id2name = {v: k for k, v in self._name2id.items()}

        if isinstance(elem_id, str):
            return id2name.get(elem_id, elem_id)
        return [self.get_name_by_id(n, id2name) for n in elem_id]

    def read_attrs(self, key: str):
        key_id = self.get_id_by_name(key)
        if self._elements is not None and key_id not in self._elements:
            raise KeyError(key)
        return self.group[key_id].attrs.asdict()

    def keys(self):
        return self._name2id.keys()

    def repr(self):
        return (f"{type(self).__name__}: {', '.join(self.keys())}")

    def __repr__(self):
        str_obj = f"{type(self).__name__} at path `{self._path}`"
        if len(self.keys()) > 0:
            str_obj += f" with elements: {', '.join(self.keys())}"
        return str_obj

    def __contains__(self, key: str):
        return key in self.keys() or key in self._name2id.values()

    def __iter__(self):
        yield from self.keys()

    def __len__(self):
        return len(self.keys())



import traceback
from ._utils import format_print, parse_to_str, DEFAULT_LOGGER


try:
    import dask
    dask.config.set({'dataframe.query-planning': False})
except Exception as e:
    DEFAULT_LOGGER.warning("".join(traceback.format_exception(e)))

from ._constants import (
    Species,
    Technologies,
    DefaultGroup,
    SpatialAttrs,
    ConnectorKeys,
    CombinedTechnologies,
    SubmissionElementKeys,
    ImagesSubmission,
    SegmentationSubmission,
    TrasncriptsSubmission,
    ExpressionSubmission,
)
from ._analysis import Analysis
from ._anndata import ConnectorAnnData
from ._connector import SpatialXConnector
from ._version import __version__


__ALL__ = [
    __version__,
    Analysis,
    ConnectorAnnData,
    Species,
    Technologies,
    DefaultGroup,
    SpatialAttrs,
    ConnectorKeys,
    CombinedTechnologies,
    SubmissionElementKeys,
    ImagesSubmission,
    SegmentationSubmission,
    TrasncriptsSubmission,
    ExpressionSubmission,
    SpatialXConnector,
    format_print,
    parse_to_str,
    DEFAULT_LOGGER,
]



from enum import Enum


V1_API: str = "pyapi/v1"
OPEN_API: str = "openapi"
TOKEN_KEY: str = "SpatialX-Token"
UCC: str = "ucc.go"

# PyAPI
PARSE_DATA_URL: str = "submission/parse_data_information"
RENAME_IMAGE_CHANNELS: str = "elements/rename_image_channels"
SIMPLIFY_SEGMENTATION: str = "elements/simplify_segmentation"

IMPORT_METADATA: str = "metadata/import_metadata"
LIST_DECONVOLUTION_MODELS: str = "query/get_default_deconvolution_models"

# OpenAPI
INFO_URL: str = "account/info"
GROUPS_URL: str = "account/groups"
LIST_URL: str = "directory/entities"
EXTERNAL_MOUNT_URL: str = "mount/external_listing"
LIST_S3: str = "cloud/setting_list"

CREATE_STUDY_URL: str = "study/create"
LIST_STUDY_URL: str = "study/list"
LIST_PUBLIC_STUDY_URL: str = "study/list_public"
DETAIL_STUDY_URL: str = "study/detail"
UPDATE_STUDY_URL: str = "study/update"
DELETE_STUDY_URL: str = "study/stop_create_submit"

CREATE_SAMPLE_URL: str = "sample/create"
LIST_SAMPLE_URL: str = "sample/list"
DETAIL_SAMPLE_URL: str = "sample/detail"
RENAME_SAMPLE_URL: str = "sample/rename"
DELETE_SAMPLE_URL: str = "sample/delete"

DETAIL_SAMPLE_DATA_URL: str = "data/detail"
ADD_SAMPLE_DATA_URL: str = "data/create"
ADD_SAMPLE_DATA_ELEMENT_URL: str = "data/add_element"
RENAME_SAMPLE_DATA_URL: str = "data/rename"
DELETE_SAMPLE_DATA_URL: str = "data/delete"
DELETE_SAMPLE_DATA_ELEMENT_URL: str = "data/delete_element"
ADD_URL: str = "data_extend/add"

UPLOAD_FILE_URL: str = "upload/simple"
UPLOAD_CHUNK_START_URL: str = "upload/chunk/start"
UPLOAD_CHUNK_PROCESS_URL: str = "upload/chunk/process"
UPLOAD_CHUNK_MERGE_URL: str = "upload/chunk/merge"
UPLOAD_CHUNK_FINISH_URL: str = "upload/chunk/finish"
UPLOAD_FOLDER_FINISH_URL: str = "upload/folder_finish"


# Convert from Lens
STUDY_CONVERT_FROM_LENS_URL: str = "study/convert_from_lens"

# Analysis
CREATE_ANALYSIS_URL: str = "analysis/create"
GET_ANNOTATED_ELEMENTS: str = "metadata/get_annotated_element"


# Types
class Technologies(Enum):
    COSMX_VER1 = "COSMX_VER1"
    COSMX_VER2 = "COSMX_VER2"
    MERSCOPE_VER1 = "MERSCOPE_VER1"
    MERSCOPE_VER2 = "MERSCOPE_VER2"
    XENIUM = "XENIUM"
    XENIUM_HE = "XENIUM_HE"

    VISIUM = "VISIUM"
    VISIUM_HD = "VISIUM_HD"
    SLIDE_SEQ = "SLIDE_SEQ"
    VISIUM_HD_TO_SC = "VISIUM_HD_TO_SC"

    SPATIALDATA_ZARR = "SPATIALDATA_ZARR"
    SPATIALDATA_H5AD = "SPATIALDATA_H5AD"
    SPATIALDATA_RDS = "SPATIALDATA_RDS"

    PROTEIN_OME_TIFF = "PROTEIN_OME_TIFF"
    PROTEIN_QPTIFF = "PROTEIN_QPTIFF"
    PROTEIN_MCD = "PROTEIN_MCD"


class SubmissionType(Enum):
    SPATIAL_TRANSCRIPTOMICS = "SUBMIT_SPATIAL_TRANSCRIPTOMICS"
    SPATIAL_BULK = "SUBMIT_SPATIAL_BULK"
    SPATIAL_PROTEOMICS = "SUBMIT_SPATIAL_PROTEOMICS"

    # OTHERS
    PROCESSED_SPATIAL_DATA = "SUBMIT_PROCESSED_SPATIAL_DATA"
    OTHERS_FORMAT_DATA = "SUBMIT_OTHERS_FORMAT_DATA"
    COMBINED_DATASETS = "COMBINED_DATASETS"
    ADDING_ELEMENTS = "ADDING_ELEMENTS"

    @classmethod
    def detect_submission_type(cls, technology: str):
        if technology in [
            Technologies.COSMX_VER1,
            Technologies.COSMX_VER1.value,
            Technologies.COSMX_VER2,
            Technologies.COSMX_VER2.value,
            Technologies.MERSCOPE_VER1,
            Technologies.MERSCOPE_VER1.value,
            Technologies.MERSCOPE_VER2,
            Technologies.MERSCOPE_VER2.value,
            Technologies.XENIUM,
            Technologies.XENIUM.value,
            Technologies.XENIUM_HE,
            Technologies.XENIUM_HE.value,
        ]:
            return cls.SPATIAL_TRANSCRIPTOMICS
        elif technology in [
            Technologies.VISIUM,
            Technologies.VISIUM.value,
            Technologies.VISIUM_HD,
            Technologies.VISIUM_HD.value,
            Technologies.SLIDE_SEQ,
            Technologies.SLIDE_SEQ.value,
        ]:
            return cls.SPATIAL_BULK
        elif technology in [
            Technologies.PROTEIN_OME_TIFF,
            Technologies.PROTEIN_OME_TIFF.value,
            Technologies.PROTEIN_QPTIFF,
            Technologies.PROTEIN_QPTIFF.value,
            Technologies.PROTEIN_MCD,
            Technologies.PROTEIN_MCD.value,
        ]:
            return cls.SPATIAL_PROTEOMICS
        elif technology in [
            Technologies.SPATIALDATA_ZARR,
            Technologies.SPATIALDATA_ZARR.value,
            Technologies.SPATIALDATA_H5AD,
            Technologies.SPATIALDATA_H5AD.value,
            Technologies.SPATIALDATA_RDS,
            Technologies.SPATIALDATA_RDS.value,
        ]:
            return cls.PROCESSED_SPATIAL_DATA
        return cls.OTHERS_FORMAT_DATA


class Species(Enum):
    HUMAN = "human"
    MOUSE = "mouse"
    OTHERS = "others"
    NON_HUMAN_PRIMATE = "nonHumanPrimate"


class StudyType(Enum):
    SINGLECELL_STUDY_TYPE_NUMBER = 0
    SPATIAL_STUDY_TYPE_NUMBER = 1


class StudyStatus(Enum):
    CREATED_STATUS = 0
    SUCCESS_STATUS = 1
    PROCESSING_STATUS = 2
    DELETE_STATUS = 3


class StudyFilter(Enum):
    EQUAL = 0
    NOT_LARGER = 1
    LARGER = 2


class DefaultGroup(Enum):
    PERSONAL_WORKSPACE = "Personal workspace"
    ALL_MEMBERS = "All members"

    LENS_GROUP_ID_PERSONAL_WORKSPACE = "personal"
    LENS_GROUP_ID_ALL_MEMBERS = "all_members"


class ImagesSubmission(Enum):
    # Transcriptomics Technologies
    COSMX_VER1 = "IMAGES_COSMX_VER1"
    COSMX_VER2 = "IMAGES_COSMX_VER2"
    MERSCOPE_VER1 = "IMAGES_MERSCOPE_VER1"
    MERSCOPE_VER2 = "IMAGES_MERSCOPE_VER2"
    XENIUM = "IMAGES_XENIUM"
    XENIUM_HE = "IMAGES_XENIUM_HE"

    # Bulk Technologies
    VISIUM = "IMAGES_VISIUM"
    VISIUM_HD = "IMAGES_VISIUM_HD"

    # Proteomics Technologies
    PROTEIN_OME_TIFF = "IMAGES_PROTEIN_OME_TIFF"  # Lunaphore COMET
    PROTEIN_QPTIFF = "IMAGES_PROTEIN_QPTIFF"  # Akoya PhenoCycler CODEX
    PROTEIN_MCD = "IMAGES_PROTEIN_MCD"  # MCD Public STTARR
    PROTEIN_COSMX_VER2 = "IMAGES_PROTEIN_COSMX_VER2"

    # Processed Data
    ZARR_SPATIALDATA = "IMAGES_ZARR_SPATIALDATA"

    # Specific cases
    TIFFFILE = "IMAGES_TIFFFILE"
    TIFFFILE_3D = "IMAGES_TIFFFILE_3D"
    TIFFFILE_HE = "IMAGES_TIFFFILE_HE"
    FROM_EXISTED = "IMAGES_FROM_EXISTED"
    PROTEIN_FROM_EXISTED = "IMAGES_PROTEIN_FROM_EXISTED"


class SegmentationSubmission(Enum):
    # Transcriptomics Technologies
    COSMX_VER1 = "SEGMENTATION_COSMX_VER1"
    COSMX_VER2 = "SEGMENTATION_COSMX_VER2"
    MERSCOPE_VER1 = "SEGMENTATION_MERSCOPE_VER1"
    MERSCOPE_VER2 = "SEGMENTATION_MERSCOPE_VER2"
    XENIUM = "SEGMENTATION_XENIUM"
    XENIUM_HE = "SEGMENTATION_XENIUM_HE"

    # Bulk Technologies
    VISIUM = "SPOT_VISIUM"
    VISIUM_HD = "SPOT_VISIUM_HD"
    SLIDE_SEQ = "SPOT_SLIDE_SEQ"
    STOMICS_BINS = "SEGMENTATION_STOMICS_BINS"
    STOMICS_CELL_BINS = "SEGMENTATION_STOMICS_CELL_BINS"

    # Processed Data
    ZARR_SPATIALDATA = "SEGMENTATION_ZARR_SPATIALDATA"

    # Specific cases
    PARQUET = "SEGMENTATION_PARQUET"
    GEOJSON = "SEGMENTATION_GEOJSON"
    FEATHER = "SEGMENTATION_FEATHER"
    HALO = "SEGMENTATION_HALO"
    CELL_MASKS = "SEGMENTATION_CELL_MASKS"


class TrasncriptsSubmission(Enum):
    # Transcriptomics Technologies
    COSMX_VER1 = "TRANSCRIPTS_COSMX_VER1"
    COSMX_VER2 = "TRANSCRIPTS_COSMX_VER2"
    MERSCOPE_VER1 = "TRANSCRIPTS_MERSCOPE_VER1"
    MERSCOPE_VER2 = "TRANSCRIPTS_MERSCOPE_VER2"
    XENIUM = "TRANSCRIPTS_XENIUM"
    XENIUM_HE = "TRANSCRIPTS_XENIUM_HE"

    # Processed Data
    ZARR_SPATIALDATA = "TRANSCRIPTS_ZARR_SPATIALDATA"

    # Specific cases
    DATAFRAME = "TRANSCRIPTS_DATAFRAME"


class ExpressionSubmission(Enum):
    IMPORT_ANNDATA = "IMPORT_ANNDATA"

    # Bulk
    VISIUM = "ANNOTATED_DATA_VISIUM"
    VISIUM_HD = "ANNOTATED_DATA_VISIUM_HD"
    SLIDE_SEQ = "ANNOTATED_DATA_SLIDE_SEQ"
    GEOMX = "ANNOTATED_DATA_GEOMX"
    STOMICS_BINS = "ANNOTATED_DATA_STOMICS_BINS"
    STOMICS_CELL_BINS = "ANNOTATED_DATA_STOMICS_CELL_BINS"

    # Processed data
    MERGE_MULTIPLE_ANNDATA = "MERGE_MULTIPLE_ANNDATA"


class CombinedTechnologies(Enum):
    VISIUM = "Visium"
    VISIUM_HD = "Visium HD"
    GEOMX = "GeoMx"
    SINGLE_CELL_TRANSCRIPTOMICS = "Single cell transcriptomics"
    SINGLE_CELL_PROTEOMICS = "Single cell proteomics"


class SubmissionElementKeys(Enum):
    IMAGES = "images"
    PROTEIN_IMAGES = "protein_images"
    SEGMENTATION = "segmentation"
    TRANSCRIPTS = "transcripts"
    CELL_CENTERS = "cell_centers"
    MATRIX = "matrix"
    EXPRESSION = "annotated_data"
    ALIGNMENT = "alignment"
    SCALEFACTORS = "scalefactors"
    TISSUE_POSITIONS = "tissue_positions"
    MPP = "mpp"

    IMAGES_ID = "images_id"
    SEGMENTATION_ID = "segmentation_id"
    SPATIAL_ID = "spatial_id"
    NUCLEI_CHANNELS = "nuclei_channels"
    MEMBRANE_CHANNELS = "membrane_channels"

    REFERENCE_ANNDATA_PATHS = "reference_anndata_paths"
    INCLUDED_IMAGES = "included_images"
    INCLUDED_SEGMENTATION = "included_segmentation"
    METADATA_ACTION = "metadata_action"


class ConnectorKeys(Enum):
    INFORMATION_FIELDS = ["email", "sub_dir", "name", "app_base_url", "routing_table"]

    # Response keys
    ENTITY = "entity"
    ENTITIES = "entities"
    MESSAGE = "message"
    STATUS = "status"
    STUDY = "study"
    SAMPLE = "sample"
    UNIQUE_ID = "unique_id"
    ROOT_FOLDER = "root_folder"

    FILE = "file"
    DIRECTORY = "directory"

    FILES = "files"
    FOLDERS = "folders"
    ARGS = "args"
    KWARGS = "kwargs"
    IDENTITIES = "identities"

    MAP_SUBMIT_RESULT = "map_submit_result"
    ANNOTATED_DATA = "annotated_data"
    TRANSCRIPTS = "transcripts"
    SPOT = "spot"
    MPP = "mpp"

    # Parameter keys
    OBS = "obs"
    OBSM = "obsm"
    BARCODES: str = "barcodes"
    GENES: str = "genes"

    STUDY_PATH = "study_path"
    STUDY_ID = "study_id"
    GROUP_ID = "group_id"
    SPECIES = "species"
    LIMIT = "limit"
    OFFSET = "offset"
    ACTIVE = "active"
    COMPARE = "compare"
    NAME = "name"
    DATA = "data"
    DATA_NAME = "data_name"
    SAMPLE_NAME = "sample_name"
    TITLE = "title"
    KEY = "key"
    VALUE = "value"
    TYPE = "type"
    PATH = "path"
    DATA_PATH = "data_path"
    IGNORE_HIDDEN = "ignore_hidden"
    TECHNOLOGY = "technology"
    SAMPLE_ID = "sample_id"
    SAMPLE_DATA = "sample_data"
    SUBMIT_ID = "submit_id"
    SUBMISSION_NAME = "submission_name"
    SUBMISSION_TYPE = "submission_type"
    SUBMISSION_INFO = "submission_info"
    NEED_DATA = "need_data"
    DATA_ID = "data_id"
    TABLE_ID = "table_id"
    ELEMENT_ID = "element_id"
    LENS_DATA_PATH = "lens_data_path"
    GROUPS = "groups"
    DEFAULT = "default"
    PARAMS = "params"
    DISPLAY_PARAMS = "display_params"
    SUB_TYPE = "sub_type"
    GROUP_TYPE = "group_type"
    DESCRIPTION = "description"
    ELEMENT = "element"
    NEW_NAME = "new_name"
    CLEANUP = "cleanup_flg"
    PUBLICATION = "publication"
    TAGS = "tags"
    AUTHOR = "author"
    ANALYSIS_ID = "analysis_id"

    EMAIL_ID = "email_id"
    TOTAL_DATA = "total_data"
    TOTAL_SAMPLE = "total_sample"
    ENABLE_STATUS = "enable_status"

    IMAGES_ID = "images_id"
    SEGMENTATION_ID = "segmentation_id"
    CHANNEL_NAMES = "channel_names"
    TOLERANCE = "tolerance"
    CSV_PATH = "csv_path"

    # Parameter upload keys
    UPLOAD_PARENT_IS_FILE = "parent_is_file"
    UPLOAD_CHUNK_SIZE = "chunk_size"
    UPLOAD_FILE_SIZE = "file_size"
    UPLOAD_OFFSET = "offset"
    UPLOAD_FILE_NAME = "name"
    UPLOAD_FOLDER_NAME = "folder_name"
    UPLOAD_UNIQUE_ID = "unique_id"
    UPLOAD_PATH = "path"
    UPLOAD_MOVE_TO_PARENT = "move_to_parent"
    UPLOAD_SENDING_INDEX = "sending_index"
    UPLOAD_FILE_DATA = "file"
    UPLOAD_TOTAL_CHUNK = "total"
    UPLOAD_IS_CHUNK = "is_chunk"
    UPLOAD_CHUNK_SMALL_SIZE = 1024 * 1024
    UPLOAD_CHUNK_MEDIUM_SIZE = 16 * 1024 * 1024
    UPLOAD_CHUNK_NORMAL_SIZE = 50 * 1024 * 1024
    UPLOAD_CHUNK_LARGE_SIZE = 100 * 1024 * 1024


class NormalizeMethod(Enum):
    RAW = "raw"
    LOG1P_NORMALIZE = "log1p-normalized"
    SQRT_NORMALIZE = "sqrt-normalized"


class SpatialAttrs(Enum):
    IMAGES = "images"
    SHAPES = "shapes"
    POINTS = "points"
    TABLES = "tables"
    LABELS = "labels"

    IMAGES_ELEMENT = "images"
    PROTEIN_IMAGES_ELEMENT = "protein_images"
    IMAGES_ALIGNMENT_ELEMENT = "images_alignment"
    SEGMENTATION_ELEMENT = "segmentation"
    TRANSCRIPTS_ELEMENT = "transcripts"
    CELL_CENTERS_ELEMENT = "cell_centers"
    ANNOTATED_DATA = "annotated_data"
    MULTI_LABELS_ELEMENT = "multi_labels"
    SPOT_ELEMENT = "spot"

    SPATIAL_ELEMENT_MAPPING = {
        IMAGES_ELEMENT: IMAGES,
        PROTEIN_IMAGES_ELEMENT: IMAGES,
        SEGMENTATION_ELEMENT: SHAPES,
        CELL_CENTERS_ELEMENT: POINTS,
        TRANSCRIPTS_ELEMENT: POINTS,
        ANNOTATED_DATA: TABLES,
        SPOT_ELEMENT: SHAPES,
        MULTI_LABELS_ELEMENT: LABELS,
    }

    CHANNELS_KEY = "channels"
    CHANNEL_LABEL = "label"


class EnableStatus(Enum):
    Status_0 = "Created"
    Status_1 = "Ready to use"
    Status_2 = "Processing"
    Status_3 = "Deleted"

    def get_status(status: int):
        res = getattr(EnableStatus, f"Status_{status}")
        if isinstance(res, Enum):
            res = res.value
        return res


class AnnDataAttr(Enum):
    X = "X"
    OBS = "obs"
    VAR = "var"
    OBSM = "obsm"
    VARM = "varm"
    OBSP = "obsp"
    VARP = "varp"


class AnnDataKeys(Enum):
    INDEX_KEY = "_index"


class AnnDataAxis(Enum):
    OBS = 0
    VAR = 1
